# SPDX-License-Identifier: GPL-2.0-or-later

from collections.abc import Iterable
import dbus as _dbus
from openrazer.client.fx import RazerFX as _RazerFX
from xml.etree import ElementTree as _ET
from openrazer.client.macro import RazerMacro as _RazerMacro


class RazerDevice(object):
    """
    Raw razer base device
    """
    _matrix_dimensions: tuple[int, int] | None
    _kbd_layout: str | None
    macro: _RazerMacro | None

    def __init__(self, serial: str) -> None:
        # Load up the DBus
        session_bus = _dbus.SessionBus()
        self._dbus = session_bus.get_object("org.razer", "/org/razer/device/{0}".format(serial))

        self._available_features = self._get_available_features()

        self._dbus_interfaces = {
            'device': _dbus.Interface(self._dbus, "razer.device.misc"),
            'brightness': _dbus.Interface(self._dbus, "razer.device.lighting.brightness")
        }

        self._name = str(self._dbus_interfaces['device'].getDeviceName())
        self._type = str(self._dbus_interfaces['device'].getDeviceType())
        self._fw = str(self._dbus_interfaces['device'].getFirmware())
        self._drv_version = str(self._dbus_interfaces['device'].getDriverVersion())
        self._has_dedicated_macro: bool | None = None
        self._device_image: str | None = None

        # Deprecated API, but kept for backwards compatibility
        self._urls = None

        self._vid, self._pid = self._dbus_interfaces['device'].getVidPid()

        self._serial = serial

        self._capabilities = {
            'name': True,
            'type': True,
            'firmware_version': True,
            'serial': True,
            'brightness': self._has_feature('razer.device.lighting.brightness'),

            'battery': self._has_feature('razer.device.power', 'getBattery'),
            'idle_time': self._has_feature('razer.device.power', ('getIdleTime', 'setIdleTime')),
            'low_battery_threshold': self._has_feature('razer.device.power', ('getLowBatteryThreshold', 'setLowBatteryThreshold')),
            # Deprecated, use idle_time & low_battery_threshold
            'get_idle_time': self._has_feature('razer.device.power', 'getIdleTime'),
            'set_idle_time': self._has_feature('razer.device.power', 'setIdleTime'),
            'get_low_battery_threshold': self._has_feature('razer.device.power', 'getLowBatteryThreshold'),
            'set_low_battery_threshold': self._has_feature('razer.device.power', 'setLowBatteryThreshold'),

            'macro_logic': self._has_feature('razer.device.macro'),
            'keyboard_layout': self._has_feature('razer.device.misc', 'getKeyboardLayout'),
            'game_mode_led': self._has_feature('razer.device.led.gamemode'),
            'keyswitch_optimization': self._has_feature('razer.device.misc.keyswitchoptimization', ('getKeyswitchOptimization', 'setKeyswitchOptimization')),
            'macro_mode_led': self._has_feature('razer.device.led.macromode', 'setMacroMode'),
            'macro_mode_led_effect': self._has_feature('razer.device.led.macromode', 'setMacroEffect'),
            'macro_mode_modifier': self._has_feature('razer.device.macro', 'setModeModifier'),
            'reactive_trigger': self._has_feature('razer.device.misc', 'triggerReactive'),

            'poll_rate': self._has_feature('razer.device.misc', ('getPollRate', 'setPollRate')),
            'supported_poll_rates': self._has_feature('razer.device.misc', 'getSupportedPollRates'),
            'dpi': self._has_feature('razer.device.dpi', ('getDPI', 'setDPI')),
            'dpi_stages': self._has_feature('razer.device.dpi', ('getDPIStages', 'setDPIStages')),
            'available_dpi': self._has_feature('razer.device.dpi', 'availableDPI'),
            'scroll_mode': self._has_feature('razer.device.scroll', ('getScrollMode', 'setScrollMode')),
            'scroll_acceleration': self._has_feature('razer.device.scroll', ('getScrollAcceleration', 'setScrollAcceleration')),
            'scroll_smart_reel': self._has_feature('razer.device.scroll', ('getScrollSmartReel', 'setScrollSmartReel')),

            # Default device is a chroma so lighting capabilities
            'lighting': self._has_feature('razer.device.lighting.chroma'),
            'lighting_breath_single': self._has_feature('razer.device.lighting.chroma', 'setBreathSingle'),
            'lighting_breath_dual': self._has_feature('razer.device.lighting.chroma', 'setBreathDual'),
            'lighting_breath_triple': self._has_feature('razer.device.lighting.chroma', 'setBreathTriple'),
            'lighting_breath_random': self._has_feature('razer.device.lighting.chroma', 'setBreathRandom'),
            'lighting_wave': self._has_feature('razer.device.lighting.chroma', 'setWave'),
            'lighting_wheel': self._has_feature('razer.device.lighting.chroma', 'setWheel'),
            'lighting_reactive': self._has_feature('razer.device.lighting.chroma', 'setReactive'),
            'lighting_none': self._has_feature('razer.device.lighting.chroma', 'setNone'),
            'lighting_spectrum': self._has_feature('razer.device.lighting.chroma', 'setSpectrum'),
            'lighting_static': self._has_feature('razer.device.lighting.chroma', 'setStatic'),
            'lighting_blinking': self._has_feature('razer.device.lighting.chroma', 'setBlinking'),

            'lighting_starlight_single': self._has_feature('razer.device.lighting.chroma', 'setStarlightSingle'),
            'lighting_starlight_dual': self._has_feature('razer.device.lighting.chroma', 'setStarlightDual'),
            'lighting_starlight_random': self._has_feature('razer.device.lighting.chroma', 'setStarlightRandom'),

            'lighting_ripple': self._has_feature('razer.device.lighting.custom', 'setRipple'),  # Thinking of extending custom to do more hence the key check
            'lighting_ripple_random': self._has_feature('razer.device.lighting.custom', 'setRippleRandomColour'),

            'lighting_pulsate': self._has_feature('razer.device.lighting.bw2013', 'setPulsate'),

            'lighting_led_matrix': bool(self._dbus_interfaces['device'].hasMatrix()),
            'lighting_led_single': self._has_feature('razer.device.lighting.chroma', 'setKey'),

            # Mouse lighting attrs
            'lighting_logo': self._has_feature('razer.device.lighting.logo'),
            'lighting_logo_active': self._has_feature('razer.device.lighting.logo', 'setLogoActive'),
            'lighting_logo_blinking': self._has_feature('razer.device.lighting.logo', 'setLogoBlinking'),
            'lighting_logo_brightness': self._has_feature('razer.device.lighting.logo', 'setLogoBrightness'),
            'lighting_logo_pulsate': self._has_feature('razer.device.lighting.logo', 'setLogoPulsate'),
            'lighting_logo_spectrum': self._has_feature('razer.device.lighting.logo', 'setLogoSpectrum'),
            'lighting_logo_static': self._has_feature('razer.device.lighting.logo', 'setLogoStatic'),
            'lighting_logo_none': self._has_feature('razer.device.lighting.logo', 'setLogoNone'),
            'lighting_logo_on': self._has_feature('razer.device.lighting.logo', 'setLogoOn'),
            'lighting_logo_reactive': self._has_feature('razer.device.lighting.logo', 'setLogoReactive'),
            'lighting_logo_wave': self._has_feature('razer.device.lighting.logo', 'setLogoWave'),
            'lighting_logo_breath_single': self._has_feature('razer.device.lighting.logo', 'setLogoBreathSingle'),
            'lighting_logo_breath_dual': self._has_feature('razer.device.lighting.logo', 'setLogoBreathDual'),
            'lighting_logo_breath_random': self._has_feature('razer.device.lighting.logo', 'setLogoBreathRandom'),
            'lighting_logo_breath_mono': self._has_feature('razer.device.lighting.logo', 'setLogoBreathMono'),

            'lighting_scroll': self._has_feature('razer.device.lighting.scroll'),
            'lighting_scroll_active': self._has_feature('razer.device.lighting.scroll', 'setScrollActive'),
            'lighting_scroll_blinking': self._has_feature('razer.device.lighting.scroll', 'setScrollBlinking'),
            'lighting_scroll_brightness': self._has_feature('razer.device.lighting.scroll', 'setScrollBrightness'),
            'lighting_scroll_pulsate': self._has_feature('razer.device.lighting.scroll', 'setScrollPulsate'),
            'lighting_scroll_spectrum': self._has_feature('razer.device.lighting.scroll', 'setScrollSpectrum'),
            'lighting_scroll_static': self._has_feature('razer.device.lighting.scroll', 'setScrollStatic'),
            'lighting_scroll_none': self._has_feature('razer.device.lighting.scroll', 'setScrollNone'),
            'lighting_scroll_on': self._has_feature('razer.device.lighting.scroll', 'setScrollOn'),
            'lighting_scroll_reactive': self._has_feature('razer.device.lighting.scroll', 'setScrollReactive'),
            'lighting_scroll_wave': self._has_feature('razer.device.lighting.scroll', 'setScrollWave'),
            'lighting_scroll_breath_single': self._has_feature('razer.device.lighting.scroll', 'setScrollBreathSingle'),
            'lighting_scroll_breath_dual': self._has_feature('razer.device.lighting.scroll', 'setScrollBreathDual'),
            'lighting_scroll_breath_random': self._has_feature('razer.device.lighting.scroll', 'setScrollBreathRandom'),
            'lighting_scroll_breath_mono': self._has_feature('razer.device.lighting.scroll', 'setScrollBreathMono'),

            'lighting_left': self._has_feature('razer.device.lighting.left'),
            'lighting_left_active': self._has_feature('razer.device.lighting.left', 'setLeftActive'),
            'lighting_left_brightness': self._has_feature('razer.device.lighting.left', 'setLeftBrightness'),
            'lighting_left_spectrum': self._has_feature('razer.device.lighting.left', 'setLeftSpectrum'),
            'lighting_left_static': self._has_feature('razer.device.lighting.left', 'setLeftStatic'),
            'lighting_left_none': self._has_feature('razer.device.lighting.left', 'setLeftNone'),
            'lighting_left_reactive': self._has_feature('razer.device.lighting.left', 'setLeftReactive'),
            'lighting_left_wave': self._has_feature('razer.device.lighting.left', 'setLeftWave'),
            'lighting_left_breath_single': self._has_feature('razer.device.lighting.left', 'setLeftBreathSingle'),
            'lighting_left_breath_dual': self._has_feature('razer.device.lighting.left', 'setLeftBreathDual'),
            'lighting_left_breath_random': self._has_feature('razer.device.lighting.left', 'setLeftBreathRandom'),
            'lighting_left_breath_mono': self._has_feature('razer.device.lighting.left', 'setLeftBreathMono'),

            'lighting_right': self._has_feature('razer.device.lighting.right'),
            'lighting_right_active': self._has_feature('razer.device.lighting.right', 'setRightActive'),
            'lighting_right_brightness': self._has_feature('razer.device.lighting.right', 'setRightBrightness'),
            'lighting_right_spectrum': self._has_feature('razer.device.lighting.right', 'setRightSpectrum'),
            'lighting_right_static': self._has_feature('razer.device.lighting.right', 'setRightStatic'),
            'lighting_right_none': self._has_feature('razer.device.lighting.right', 'setRightNone'),
            'lighting_right_reactive': self._has_feature('razer.device.lighting.right', 'setRightReactive'),
            'lighting_right_wave': self._has_feature('razer.device.lighting.right', 'setRightWave'),
            'lighting_right_breath_single': self._has_feature('razer.device.lighting.right', 'setRightBreathSingle'),
            'lighting_right_breath_dual': self._has_feature('razer.device.lighting.right', 'setRightBreathDual'),
            'lighting_right_breath_random': self._has_feature('razer.device.lighting.right', 'setRightBreathRandom'),
            'lighting_right_breath_mono': self._has_feature('razer.device.lighting.right', 'setRightBreathMono'),

            'lighting_backlight': self._has_feature('razer.device.lighting.backlight'),
            'lighting_backlight_active': self._has_feature('razer.device.lighting.backlight', 'setBacklightActive'),
            'lighting_backlight_brightness': self._has_feature('razer.device.lighting.backlight', 'setBacklightBrightness'),
            'lighting_backlight_spectrum': self._has_feature('razer.device.lighting.backlight', 'setBacklightSpectrum'),
            'lighting_backlight_static': self._has_feature('razer.device.lighting.backlight', 'setBacklightStatic'),
            'lighting_backlight_none': self._has_feature('razer.device.lighting.backlight', 'setBacklightNone'),
            'lighting_backlight_on': self._has_feature('razer.device.lighting.backlight', 'setBacklightOn'),
            'lighting_backlight_reactive': self._has_feature('razer.device.lighting.backlight', 'setBacklightReactive'),
            'lighting_backlight_wave': self._has_feature('razer.device.lighting.backlight', 'setBacklightWave'),
            'lighting_backlight_breath_single': self._has_feature('razer.device.lighting.backlight', 'setBacklightBreathSingle'),
            'lighting_backlight_breath_dual': self._has_feature('razer.device.lighting.backlight', 'setBacklightBreathDual'),
            'lighting_backlight_breath_random': self._has_feature('razer.device.lighting.backlight', 'setBacklightBreathRandom'),
            'lighting_backlight_breath_mono': self._has_feature('razer.device.lighting.backlight', 'setBacklightBreathMono'),

            'lighting_profile_led_red': self._has_feature('razer.device.lighting.profile_led', 'setRedLED'),
            'lighting_profile_led_green': self._has_feature('razer.device.lighting.profile_led', 'setGreenLED'),
            'lighting_profile_led_blue': self._has_feature('razer.device.lighting.profile_led', 'setBlueLED'),

            # Charging Pad attrs
            'lighting_charging': self._has_feature('razer.device.lighting.charging'),
            'lighting_charging_active': self._has_feature('razer.device.lighting.charging', 'setChargingActive'),
            'lighting_charging_brightness': self._has_feature('razer.device.lighting.charging', 'setChargingBrightness'),
            'lighting_charging_spectrum': self._has_feature('razer.device.lighting.charging', 'setChargingSpectrum'),
            'lighting_charging_static': self._has_feature('razer.device.lighting.charging', 'setChargingStatic'),
            'lighting_charging_none': self._has_feature('razer.device.lighting.charging', 'setChargingNone'),
            'lighting_charging_wave': self._has_feature('razer.device.lighting.charging', 'setChargingWave'),
            'lighting_charging_breath_single': self._has_feature('razer.device.lighting.charging', 'setChargingBreathSingle'),
            'lighting_charging_breath_dual': self._has_feature('razer.device.lighting.charging', 'setChargingBreathDual'),
            'lighting_charging_breath_random': self._has_feature('razer.device.lighting.charging', 'setChargingBreathRandom'),
            'lighting_charging_breath_mono': self._has_feature('razer.device.lighting.charging', 'setChargingBreathMono'),

            'lighting_fast_charging': self._has_feature('razer.device.lighting.fast_charging'),
            'lighting_fast_charging_active': self._has_feature('razer.device.lighting.fast_charging', 'setFastChargingActive'),
            'lighting_fast_charging_brightness': self._has_feature('razer.device.lighting.fast_charging', 'setFastChargingBrightness'),
            'lighting_fast_charging_spectrum': self._has_feature('razer.device.lighting.fast_charging', 'setFastChargingSpectrum'),
            'lighting_fast_charging_static': self._has_feature('razer.device.lighting.fast_charging', 'setFastChargingStatic'),
            'lighting_fast_charging_none': self._has_feature('razer.device.lighting.fast_charging', 'setFastChargingNone'),
            'lighting_fast_charging_wave': self._has_feature('razer.device.lighting.fast_charging', 'setFastChargingWave'),
            'lighting_fast_charging_breath_single': self._has_feature('razer.device.lighting.fast_charging', 'setFastChargingBreathSingle'),
            'lighting_fast_charging_breath_dual': self._has_feature('razer.device.lighting.fast_charging', 'setFastChargingBreathDual'),
            'lighting_fast_charging_breath_random': self._has_feature('razer.device.lighting.fast_charging', 'setFastChargingBreathRandom'),
            'lighting_fast_charging_breath_mono': self._has_feature('razer.device.lighting.fast_charging', 'setFastChargingBreathMono'),

            'lighting_fully_charged': self._has_feature('razer.device.lighting.fully_charged'),
            'lighting_fully_charged_active': self._has_feature('razer.device.lighting.fully_charged', 'setFullyChargedActive'),
            'lighting_fully_charged_brightness': self._has_feature('razer.device.lighting.fully_charged', 'setFullyChargedBrightness'),
            'lighting_fully_charged_spectrum': self._has_feature('razer.device.lighting.fully_charged', 'setFullyChargedSpectrum'),
            'lighting_fully_charged_static': self._has_feature('razer.device.lighting.fully_charged', 'setFullyChargedStatic'),
            'lighting_fully_charged_none': self._has_feature('razer.device.lighting.fully_charged', 'setFullyChargedNone'),
            'lighting_fully_charged_wave': self._has_feature('razer.device.lighting.fully_charged', 'setFullyChargedWave'),
            'lighting_fully_charged_breath_single': self._has_feature('razer.device.lighting.fully_charged', 'setFullyChargedBreathSingle'),
            'lighting_fully_charged_breath_dual': self._has_feature('razer.device.lighting.fully_charged', 'setFullyChargedBreathDual'),
            'lighting_fully_charged_breath_random': self._has_feature('razer.device.lighting.fully_charged', 'setFullyChargedBreathRandom'),
            'lighting_fully_charged_breath_mono': self._has_feature('razer.device.lighting.fully_charged', 'setFullyChargedBreathMono'),
        }

        # Nasty hack to convert dbus.Int32 into native
        if self.has('lighting_led_matrix'):
            self._matrix_dimensions = tuple([int(dim) for dim in self._dbus_interfaces['device'].getMatrixDimensions()])
        else:
            self._matrix_dimensions = None

        if self.has('keyboard_layout'):
            self._kbd_layout = str(self._dbus_interfaces['device'].getKeyboardLayout())
        else:
            self._kbd_layout = None

        # Setup FX
        self.fx = _RazerFX(serial, capabilities=self._capabilities, daemon_dbus=self._dbus, matrix_dims=self._matrix_dimensions)

        # Setup Macro
        if self.has('macro_logic'):
            self.macro = _RazerMacro(serial, self.name, daemon_dbus=self._dbus, capabilities=self._capabilities)
        else:
            self.macro = None

        if self.has('dpi'):
            self._dbus_interfaces['dpi'] = _dbus.Interface(self._dbus, "razer.device.dpi")
        if self.has('battery'):
            self._dbus_interfaces['power'] = _dbus.Interface(self._dbus, "razer.device.power")
        if self.has('game_mode_led'):
            self._dbus_interfaces['game_mode_led'] = _dbus.Interface(self._dbus, "razer.device.led.gamemode")
        if self.has('keyswitch_optimization'):
            self._dbus_interfaces['keyswitch_optimization'] = _dbus.Interface(self._dbus, "razer.device.misc.keyswitchoptimization")
        if self.has('macro_mode_led'):
            self._dbus_interfaces['macro_mode_led'] = _dbus.Interface(self._dbus, "razer.device.led.macromode")
        if self.has('lighting_profile_led_red') or self.has('lighting_profile_led_green') or self.has('lighting_profile_led_blue'):
            self._dbus_interfaces['profile_led'] = _dbus.Interface(self._dbus, "razer.device.lighting.profile_led")
        if self.has('scroll_mode') or self.has('scroll_acceleration') or self.has('scroll_smart_reel'):
            self._dbus_interfaces['scroll'] = _dbus.Interface(self._dbus, "razer.device.scroll")

    def _get_available_features(self) -> dict[str, list[str]]:
        introspect_interface = _dbus.Interface(self._dbus, 'org.freedesktop.DBus.Introspectable')
        xml_spec = introspect_interface.Introspect()
        root = _ET.fromstring(xml_spec)

        interfaces = {}

        for child in root:

            if child.tag != 'interface' or child.attrib.get('name') == 'org.freedesktop.DBus.Introspectable':
                continue

            current_interface = child.attrib['name']
            current_interface_methods: list[str] = []

            for method in child:
                if method.tag != 'method':
                    continue

                name = method.attrib.get('name')
                if name is None:
                    raise RuntimeError(f"Attribute name is unexpectedly None: {method.attrib}")
                current_interface_methods.append(name)

            interfaces[current_interface] = current_interface_methods

        return interfaces

    def _has_feature(self, object_path: str, method_name: str | Iterable[str] | None = None) -> bool:
        """
        Checks to see if the device has said DBus method

        :param object_path: Object path
        :type object_path: str

        :param method_name: Method name, or list of methods
        :type method_name: str or list or tuple

        :return: True if method/s exist
        :rtype: bool
        """
        if method_name is None:
            return object_path in self._available_features
        elif isinstance(method_name, str):
            return object_path in self._available_features and method_name in self._available_features[object_path]
        elif isinstance(method_name, (list, tuple)):
            result = True
            for method in method_name:
                result &= object_path in self._available_features and method in self._available_features[object_path]
            return result
        else:
            return False

    def has(self, capability: str) -> bool:
        """
        Convenience function to check capability

        :param capability: Device capability
        :type capability: str

        :return: True or False
        :rtype: bool
        """
        # Could do capability in self._capabilitys but they might be explicitly disabled
        return self._capabilities.get(capability, False)

    @property
    def name(self) -> str:
        """
        Device name

        :return: Device Name
        :rtype: str
        """
        return self._name

    @property
    def type(self) -> str:
        """
        Get device type

        :return: Device Type
        :rtype: str
        """
        return self._type

    @property
    def firmware_version(self) -> str:
        """
        Device's firmware version

        :return: FW Version
        :rtype: str
        """
        return self._fw

    @property
    def driver_version(self) -> str:
        """
        Device's driver version

        :return: Driver Version
        :rtype: str
        """
        return self._drv_version

    @property
    def serial(self) -> str:
        """
        Device's serial

        :return: Device Serial
        :rtype: str
        """
        return self._serial

    @property
    def keyboard_layout(self) -> str:
        """
        Device's keyboard layout

        :return: Keyboard layout
        :rtype: str
        """
        if self._kbd_layout:
            return self._kbd_layout
        else:
            raise NotImplementedError()

    @property
    def brightness(self) -> float:
        """
        Get device brightness

        :return: Device brightness
        :rtype: float
        """
        return float(self._dbus_interfaces['brightness'].getBrightness())

    @brightness.setter
    def brightness(self, value: float) -> None:
        """
        Set device brightness

        :param value: Device brightness
        :type value: float

        :raises ValueError: When brightness is not a float or not in range 0.0->100.0
        """
        if isinstance(value, int):
            value = float(value)

        if not isinstance(value, float):
            raise ValueError("Brightness must be a float")

        if value < 0.0 or value > 100.0:
            raise ValueError("Brightness must be between 0 and 100")

        self._dbus_interfaces['brightness'].setBrightness(value)

    @property
    def capabilities(self) -> dict[str, bool]:
        """
        Device capabilities

        :return: Device capabilities
        :rtype: dict
        """
        return self._capabilities

    @property
    def dedicated_macro(self) -> bool:
        """
        Device has dedicated macro keys

        :return: If the device has macro keys
        :rtype: bool
        """
        if self._has_dedicated_macro is None:
            self._has_dedicated_macro = self._dbus_interfaces['device'].hasDedicatedMacroKeys()

        return self._has_dedicated_macro

    @property
    def device_image(self) -> str:
        if self._device_image is None:
            self._device_image = str(self._dbus_interfaces['device'].getDeviceImage())

        return self._device_image

    @property
    def razer_urls(self) -> dict[str, bool | str]:
        # Deprecated API, but kept for backwards compatibility
        return {
            "DEPRECATED": True,
            "top_img": self.device_image,
            "side_img": self.device_image,
            "perspective_img": self.device_image
        }

    @property
    def battery_level(self) -> int:
        """
        Get battery level from device

        :return: Battery level (0-100)
        """
        if self.has('battery'):
            return int(self._dbus_interfaces['power'].getBattery())
        else:
            raise NotImplementedError()

    @property
    def is_charging(self) -> bool:
        """
        Get whether the device is charging or not

        :return: Boolean
        """
        if self.has('battery'):
            return bool(self._dbus_interfaces['power'].isCharging())
        else:
            raise NotImplementedError()

    def set_idle_time(self, idle_time: int) -> None:
        """
        Sets the idle time on the device

        :param idle_time: the time in seconds
        """
        if self.has('battery'):
            self._dbus_interfaces['power'].setIdleTime(idle_time)
        else:
            raise NotImplementedError()

    def get_idle_time(self) -> int:
        """
        Gets the idle time of the device

        :return: Number of seconds before this device goes into powersave
                 (60-900)
        """
        if self.has('battery'):
            return int(self._dbus_interfaces['power'].getIdleTime())
        else:
            raise NotImplementedError()

    def set_low_battery_threshold(self, threshold: int) -> None:
        """
        Set the low battery threshold as a percentage

        :param threshold: Battery threshold as a percentage
        :type threshold: int
        """
        if self.has('battery'):
            self._dbus_interfaces['power'].setLowBatteryThreshold(threshold)
        else:
            raise NotImplementedError()

    def get_low_battery_threshold(self) -> int:
        """
        Get the low battery threshold as a percentage

        :return: Battery threshold as a percentage
        """
        if self.has('battery'):
            return int(self._dbus_interfaces['power'].getLowBatteryThreshold())
        else:
            raise NotImplementedError()

    @property
    def poll_rate(self) -> int:
        """
        Get poll rate from device

        :return: Poll rate
        :rtype: int

        :raises NotImplementedError: If function is not supported
        """
        if self.has('poll_rate'):
            return int(self._dbus_interfaces['device'].getPollRate())
        else:
            raise NotImplementedError()

    @poll_rate.setter
    def poll_rate(self, poll_rate: int) -> None:
        """
        Set poll rate of device

        :param poll_rate: Polling rate
        :type poll_rate: int

        :raises NotImplementedError: If function is not supported
        """
        if self.has('poll_rate'):
            if not isinstance(poll_rate, int):
                raise ValueError("Poll rate is not an integer: {0}".format(poll_rate))

            self._dbus_interfaces['device'].setPollRate(poll_rate)

        else:
            raise NotImplementedError()

    @property
    def supported_poll_rates(self) -> list[int]:
        """
        Get poll rates supported by the device

        :return: Supported poll rates
        :rtype: list

        :raises NotImplementedError: If function is not supported
        """
        if self.has('supported_poll_rates'):
            dbuslist = self._dbus_interfaces['device'].getSupportedPollRates()
            # Repack list from dbus ints to normal ints
            return [int(d) for d in dbuslist]
        else:
            raise NotImplementedError()

    def __str__(self) -> str:
        return self._name

    def __repr__(self) -> str:
        return '<{0} {1}>'.format(self.__class__.__name__, self._serial)


class BaseDeviceFactory(object):
    @staticmethod
    def get_device(serial: str) -> RazerDevice:
        raise NotImplementedError()



# SPDX-License-Identifier: GPL-2.0-or-later

"""
Class to hold a device and collections of them
"""


class Device(object):
    """
    Razer Device (High level not dbus)
    """

    def __init__(self, device_id, device_serial, device_dbus_object):
        self._parent = None

        self._id = device_id
        self._serial = device_serial
        self._dbus = device_dbus_object
        # Register as parent
        self._dbus.register_parent(self)

    @property
    def device_id(self):
        """
        Device's USB ID String

        :return: Device ID
        :rtype: str
        """
        return self._id

    @property
    def serial(self):
        """
        Device's Serial String

        :return: Serial
        :rtype: str
        """
        return self._serial

    @property
    def dbus(self):
        """
        Device's DBus object

        :return: DBus Object
        :rtype: openrazer_daemon.hardware.device_base.__RazerDevice
        """
        return self._dbus

    def register_parent(self, parent):
        """
        Register the parent as an observer to be optionally notified (sends to other devices)

        :param parent: Observer
        :type parent: object
        """
        self._parent = parent

    def notify_parent(self, msg):
        """
        Notify observers with msg

        :param msg: Tuple with first element a string
        :type msg: tuple
        """
        self._parent.notify(self, msg)

    def notify_child(self, msg):
        """
        Receive observer messages

        :param msg: Tuple with first element a string
        :type msg: tuple
        """
        # Message from DBus object
        self._dbus.notify(msg)


class DeviceCollection(object):
    """
    Multimap of devices

    Can be referenced by either ID or serial
    """

    def __init__(self):
        self._id_map = {}
        self._serial_map = {}

    def add(self, device_id, device_serial, device_dbus):
        """
        Add device to collection

        :param device_id: Device's USB ID
        :type device_id: str

        :param device_serial: Device's Serial String
        :type device_serial: str

        :param device_dbus: Device's DBus object
        :type device_dbus: openrazer_daemon.hardware.device_base.__RazerDevice
        """
        device_object = Device(device_id, device_serial, device_dbus)
        device_object.register_parent(self)

        self._id_map[device_id] = device_object
        self._serial_map[device_serial] = device_object

    def remove(self, key):
        """
        Remove object being referenced to by ident from collection

        :param key: ID or serial
        :type key: str
        """
        self.__delitem__(key)

    def get(self, item):
        """
        Get device object by ID or serial

        :param item: ID or serial
        :type item: str

        :return: Device object
        :rtype: Device

        :raises IndexError: If item not found
        """
        return self.__getitem__(item)

    def id_items(self):
        """
        Get (id, Device) iterator

        :return: Items method from the id map
        :rtype: list of tuple of str, Device
        """
        return self._id_map.items()

    def serial_items(self):
        """
        Get (serial, Device) iterator

        :return: Items method from the serial map
        :rtype: list of tuple of str, Device
        """
        return self._serial_map.items()

    def serials(self):
        """
        Get list of serials

        :return: Serial list
        :rtype: list of str
        """
        return list(self._serial_map.keys())

    def __len__(self):
        """
        Get length of collection

        :return: Length
        :rtype: int
        """
        return len(self._id_map)

    def __getitem__(self, item):
        """
        Get device object by ID or serial

        :param item: ID or serial
        :type item: str

        :return: Device object
        :rtype: Device

        :raises IndexError: If item not found
        """
        if item in self._id_map:
            return self._id_map[item]
        elif item in self._serial_map:
            return self._serial_map[item]
        else:
            raise IndexError()

    def __delitem__(self, key):
        """
        Remove object being referenced to by ident from collection

        :param key: ID or serial
        :type key: str
        """
        if key in self._id_map:
            serial = self._id_map[key].serial
            self._id_map.pop(key, None)
            self._serial_map.pop(serial, None)
        elif key in self._serial_map:
            device_id = self._serial_map[key].device_id
            self._id_map.pop(device_id, None)
            self._serial_map.pop(key, None)

    def __contains__(self, item):
        """
        If ID or serial exists in datastructure

        :param item: ID or serial
        :type item: str

        :return: True if ID or serial exists
        :rtype: bool
        """
        return item in self._id_map or item in self._serial_map

    def __iter__(self):
        """
        Get devices

        :return: Devices
        :rtype: list of Device
        """
        return iter(self._id_map.values())

    @property
    def devices(self):
        """
        Get device list

        :return: List of devices
        :rtype: list of Device
        """
        return list(self._id_map.values())

    def notify(self, active_child, msg):
        """
        Send messages between children

        :param active_child: Child sending the message
        :type active_child: Device

        :param msg: Messgae
        :type msg: tuple
        """
        for child in self._id_map.values():
            if child is not active_child:
                child.notify_child(msg)



#!/usr/bin/python3
import argparse
import glob
import os
import time
import random


COLOURS = (b'\xFF\x00\x00', b'\x00\xFF\x00', b'\x00\x00\xFF', b'\xFF\xFF\x00', b'\xFF\x00\xFF', b'\x00\xFF\xFF')


def write_binary(driver_path, device_file, payload):
    with open(os.path.join(driver_path, device_file), 'wb') as open_file:
        open_file.write(payload)


def read_string(driver_path, device_file):
    with open(os.path.join(driver_path, device_file), 'r') as open_file:
        return open_file.read().rstrip('\n')


def write_string(driver_path, device_file, payload):
    with open(os.path.join(driver_path, device_file), 'w') as open_file:
        open_file.write(payload)


def find_devices(vid, pid):
    driver_paths = glob.glob(os.path.join('/sys/bus/hid/drivers/razeraccessory', '*:{0:04X}:{1:04X}.*'.format(vid, pid)))

    for driver_path in driver_paths:
        device_type_path = os.path.join(driver_path, 'device_type')

        if os.path.exists(device_type_path):
            yield driver_path


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('--skip-standard', action='store_true')
    parser.add_argument('--skip-custom', action='store_true')
    parser.add_argument('--skip-game-led', action='store_true')
    parser.add_argument('--skip-macro-led', action='store_true')

    return parser.parse_args()


if __name__ == '__main__':
    args = parse_args()

    found_chroma = False

    for index, driver_path in enumerate(find_devices(0x1532, 0x0068), start=1):
        found_chroma = True

        print("Razer Firefly_hyperflux {0}\n".format(index))

        print("Driver version: {0}".format(read_string(driver_path, 'version')))
        print("Driver firmware version: {0}".format(read_string(driver_path, 'firmware_version')))
        print("Device serial: {0}".format(read_string(driver_path, 'device_serial')))
        print("Device type: {0}".format(read_string(driver_path, 'device_type')))
        print("Device mode: {0}".format(read_string(driver_path, 'device_mode')))

        # Set to static red so that we have something standard
        write_binary(driver_path, 'matrix_effect_static', b'\xFF\x00\x00')

        if not args.skip_standard:
            print("Starting brightness test. Press enter to begin.")
            input()
            print("Max brightness...", end='')
            write_string(driver_path, 'matrix_brightness', '255')
            time.sleep(1)
            print("brightness ({0})".format(read_string(driver_path, 'matrix_brightness')))
            time.sleep(1)
            print("Half brightness...", end='')
            write_string(driver_path, 'matrix_brightness', '128')
            time.sleep(1)
            print("brightness ({0})".format(read_string(driver_path, 'matrix_brightness')))
            time.sleep(1)
            print("Zero brightness...", end='')
            write_string(driver_path, 'matrix_brightness', '0')
            time.sleep(1)
            print("brightness ({0})".format(read_string(driver_path, 'matrix_brightness')))
            time.sleep(1)
            write_string(driver_path, 'matrix_brightness', '255')

            print("Starting other colour effect tests. Press enter to begin.")
            input()
            print("Green Static")
            write_binary(driver_path, 'matrix_effect_static', b'\x00\xFF\x00')
            time.sleep(5)
            print("Cyan Static")
            write_binary(driver_path, 'matrix_effect_static', b'\x00\xFF\xFF')
            time.sleep(5)
            print("Spectrum")
            write_binary(driver_path, 'matrix_effect_spectrum', b'\x00')
            time.sleep(10)
            print("None")
            write_binary(driver_path, 'matrix_effect_none', b'\x00')
            time.sleep(5)
            print("Wave Left")
            write_string(driver_path, 'matrix_effect_wave', '1')
            time.sleep(5)
            print("Wave Right")
            write_string(driver_path, 'matrix_effect_wave', '2')
            time.sleep(5)
            print("Breathing random")
            write_binary(driver_path, 'matrix_effect_breath', b'\x00')
            time.sleep(10)
            print("Breathing red")
            write_binary(driver_path, 'matrix_effect_breath', b'\xFF\x00\x00')
            time.sleep(10)
            print("Breathing blue-green")
            write_binary(driver_path, 'matrix_effect_breath', b'\x00\xFF\x00\x00\x00\xFF')
            time.sleep(10)

        if not args.skip_custom:
            # Custom LEDs all rows
            payload_all = b'\x00\x00\x0E'
            for i in range(0, 15):  # 15 colours 0x00-0x0E
                payload_all += random.choice(COLOURS)

            payload_m1_5 = b''
            for led in (0x00, 0x0E):
                led_byte = led.to_bytes(1, byteorder='big')
                payload_m1_5 += b'\x00' + led_byte + led_byte + b'\xFF\xFF\xFF'

            print("Custom LED matrix colours test. Press enter to begin.")
            input()
            write_binary(driver_path, 'matrix_custom_frame', payload_all)
            write_binary(driver_path, 'matrix_effect_custom', b'\x00')
            print("Custom LED matrix partial colours test. First and last led to white. Press enter to begin.")
            input()
            write_binary(driver_path, 'matrix_custom_frame', payload_m1_5)
            write_binary(driver_path, 'matrix_effect_custom', b'\x00')
            time.sleep(0.5)

        print("Finished")

    if not found_chroma:
        print("No Fireflies found")



import sys
import argparse

import openrazer.client

choices = ('get', 'on', 'off', 'flashing', 'static', 'effect_state')
parser = argparse.ArgumentParser()
parser.add_argument('action', choices=choices)
args = parser.parse_args()

device_manager = openrazer.client.DeviceManager()
keyboard = None

for device in device_manager.devices:
    if device.type == 'keyboard':
        keyboard = device
        break
else:
    print("Could not find suitable keyboard", file=sys.stderr)
    sys.exit(1)


# Check keyboard has macro_mode
if not keyboard.has('macro_mode_led'):
    print("Keyboard doesn't have macro mode", file=sys.stderr)
    sys.exit(1)


if args.action == 'get':
    if keyboard.macro_mode_led:
        print("Macro mode LED: On")
    else:
        print("Macro mode LED: Off")

elif args.action == 'effect_state':
    if keyboard.macro_mode_led_effect == openrazer.client.constants.MACRO_LED_BLINK:
        print("Macro mode LED: Blinking")
    else:
        print("Macro mode LED: Static")

elif args.action == 'on':
    keyboard.macro_mode_led = True
    print("Turned macro mode LED on")

elif args.action == 'flashing':
    keyboard.macro_mode_led = True
    keyboard.macro_mode_led_effect = openrazer.client.constants.MACRO_LED_BLINK
    print("Turned macro mode LED on (flashing)")

elif args.action == 'static':
    keyboard.macro_mode_led = True
    keyboard.macro_mode_led_effect = openrazer.client.constants.MACRO_LED_STATIC
    print("Turned macro mode LED on (static)")

else:  # Off
    keyboard.macro_mode_led = False
    print("Turned macro mode LED off")



# SPDX-License-Identifier: GPL-2.0-or-later

from openrazer.client.devices import RazerDevice as __RazerDevice


class RazerMouse(__RazerDevice):
    @property
    def max_dpi(self) -> int:
        """
        Gets max DPI

        :return: Max DPI, if device does not have DPI it'll raise an exception
        :rtype: int
        """
        if self.has('dpi'):
            return int(self._dbus_interfaces['dpi'].maxDPI())
        else:
            raise NotImplementedError()

    @property
    def available_dpi(self) -> list[int]:
        """
        Gets the available DPI

        :return: Available DPI, if device has only a couple of fixed possible DPI values
        :rtype: list
        """
        if self.has('available_dpi'):
            dbuslist = self._dbus_interfaces['dpi'].availableDPI()
            # Repack list from dbus ints to normal ints
            return [int(d) for d in dbuslist]
        else:
            raise NotImplementedError()

    @property
    def dpi(self) -> tuple[int, int]:
        """
        Get mouse DPI

        Will return a tuple
        :return: DPI (500, 500)
        :rtype: tuple

        :raises NotImplementedError: If function is not supported
        """
        if self.has('available_dpi'):
            dpi_x = self._dbus_interfaces['dpi'].getDPI()[0]
            return int(dpi_x), 0
        elif self.has('dpi'):
            dpi_x, dpi_y = self._dbus_interfaces['dpi'].getDPI()
            # Converting to integers to remove the dbus types
            return int(dpi_x), int(dpi_y)
        else:
            raise NotImplementedError()

    @dpi.setter
    def dpi(self, value: tuple[int, int]) -> None:
        """
        Set mouse dpi

        Daemon does type validation but can't be too careful
        :param value: DPI X, Y tuple
        :type value: tuple

        :raises ValueError: If the tuple isn't long enough or contains invalid crap
        :raises NotImplementedError: If function is not supported
        """
        if self.has('dpi'):
            if len(value) != 2:
                raise ValueError("DPI tuple is not of length 2. Length: {0}".format(len(value)))
            max_dpi = self.max_dpi
            dpi_x, dpi_y = value
            dpi_x_only = self.has('available_dpi')

            if not isinstance(dpi_x, int) or not isinstance(dpi_y, int):
                raise ValueError("DPI X or Y is not an integer, X:{0} Y:{1}".format(type(dpi_x), type(dpi_y)))

            if dpi_x < 0 or dpi_x > max_dpi:
                raise ValueError("DPI X either too small or too large, X:{0}".format(dpi_x))

            if dpi_x_only and not dpi_y == 0:
                raise ValueError("DPI Y is not supported for this device")
            elif dpi_y < 0 or dpi_y > max_dpi:
                raise ValueError("DPI Y either too small or too large, Y:{0}".format(dpi_y))

            self._dbus_interfaces['dpi'].setDPI(dpi_x, dpi_y)
        else:
            raise NotImplementedError()

    @property
    def dpi_stages(self) -> tuple[int, list[tuple[int, int]]]:
        """
        Get mouse DPI stages

        Will return a tuple containing the active DPI stage number and the list
        of DPI stages as tuples.
        The active DPI stage number must be: >= 1 and <= nr of DPI stages.
        :return: active DPI stage number and DPI stages
                 (1, [(500, 500), (1000, 1000), (2000, 2000) ...]
        :rtype: (int, list)

        :raises NotImplementedError: if function is not supported
        """
        if self.has('dpi_stages'):
            response = self._dbus_interfaces['dpi'].getDPIStages()
            dpi_stages = []

            active_stage = int(response[0])

            for dpi_x, dpi_y in response[1]:
                dpi_stages.append((int(dpi_x), int(dpi_y)))

            return (active_stage, dpi_stages)
        else:
            raise NotImplementedError()

    @dpi_stages.setter
    def dpi_stages(self, value: tuple[int, list[tuple[int, int]]]) -> None:
        """
        Set mouse DPI stages

        Daemon does type validation but can't be too careful
        :param value: active DPI stage number and list of DPI X, Y tuples
        :type value: (int, list)

        :raises ValueError: when the input is invalid
        :raises NotImplementedError: If function is not supported
        """
        if self.has('dpi_stages'):
            max_dpi = self.max_dpi
            dpi_stages = []

            active_stage = value[0]
            if not isinstance(active_stage, int):
                raise ValueError(
                    "Active DPI stage is not an integer: {0}".format(
                        type(active_stage)))

            if active_stage < 1:
                raise ValueError(
                    "Active DPI stage has invalid value: {0} < 1".format(
                        active_stage))

            for stage in value[1]:
                if len(stage) != 2:
                    raise ValueError(
                        "DPI tuple is not of length 2. Length: {0}".format(
                            len(stage)))

                dpi_x, dpi_y = stage

                if not isinstance(dpi_x, int) or not isinstance(dpi_y, int):
                    raise ValueError(
                        "DPI X or Y is not an integer, X:{0} Y:{1}".format(
                            type(dpi_x), type(dpi_y)))

                if dpi_x < 0 or dpi_x > max_dpi:
                    raise ValueError(
                        "DPI X either too small or too large, X:{0}".format(
                            dpi_x))
                if dpi_y < 0 or dpi_y > max_dpi:
                    raise ValueError(
                        "DPI Y either too small or too large, Y:{0}".format(
                            dpi_y))

                dpi_stages.append((dpi_x, dpi_y))

            if active_stage > len(dpi_stages):
                raise ValueError(
                    "Active DPI stage has invalid value: {0} > {1}".format(
                        active_stage, len(dpi_stages)))

            self._dbus_interfaces['dpi'].setDPIStages(active_stage, dpi_stages)
        else:
            raise NotImplementedError()

    @property
    def scroll_mode(self) -> int:
        """
        Get the scroll wheel mode of the device

        :return: The device's current scroll mode (0 = tactile, 1 = free spin)
        :rtype: int

        :raises NotImplementedError: If function is not supported
        """
        if self.has('scroll_mode'):
            return int(self._dbus_interfaces['scroll'].getScrollMode())
        else:
            raise NotImplementedError()

    @scroll_mode.setter
    def scroll_mode(self, mode: int) -> None:
        """
        Set the scroll mode of the device

        :param mode: The mode to set (0 = tactile, 1 = free spin)
        :type mode: int

        :raises NotImplementedError: If function is not supported
        """
        if self.has('scroll_mode'):
            self._dbus_interfaces['scroll'].setScrollMode(mode)
        else:
            raise NotImplementedError()

    @property
    def scroll_acceleration(self) -> bool:
        """
        Get the device's scroll acceleration state

        :return: true if acceleration enabled, false otherwise
        :rtype: bool

        :raises NotImplementedError: If function is not supported
        """
        if self.has('scroll_acceleration'):
            return bool(self._dbus_interfaces['scroll'].getScrollAcceleration())
        else:
            raise NotImplementedError()

    @scroll_acceleration.setter
    def scroll_acceleration(self, enabled: bool) -> None:
        """
        Set the device's scroll acceleration state

        :param enabled: true to enable acceleration, false to disable it
        :type enabled: bool

        :raises NotImplementedError: If function is not supported
        """
        if self.has('scroll_acceleration'):
            self._dbus_interfaces['scroll'].setScrollAcceleration(enabled)
        else:
            raise NotImplementedError()

    @property
    def scroll_smart_reel(self) -> bool:
        """
        Get the device's "smart reel" state

        :return: true if smart reel enabled, false otherwise
        :rtype: bool

        :raises NotImplementedError: If function is not supported
        """
        if self.has('scroll_smart_reel'):
            return bool(self._dbus_interfaces['scroll'].getScrollSmartReel())
        else:
            raise NotImplementedError()

    @scroll_smart_reel.setter
    def scroll_smart_reel(self, enabled: bool) -> None:
        """
        Set the device's "smart reel" state

        :param enabled: true to enable smart reel, false to disable it
        :type enabled: bool

        :raises NotImplementedError: If function is not supported
        """
        if self.has('scroll_smart_reel'):
            self._dbus_interfaces['scroll'].setScrollSmartReel(enabled)
        else:
            raise NotImplementedError()



from openrazer.client import DeviceManager

# Create a DeviceManager. This is used to get specific devices
device_manager = DeviceManager()


print("Found {} Razer devices".format(len(device_manager.devices)))
print()


# Iterate over each device and pretty out some standard information about each
for device in device_manager.devices:
    print("{}:".format(device.name))
    print("   type: {}".format(device.type))
    print("   serial: {}".format(device.serial))
    print("   firmware version: {}".format(device.firmware_version))
    print("   driver version: {}".format(device.driver_version))
    print()



import os
import sys
import argparse
import struct
from openrazer_daemon.keyboard import EVENT_MAPPING, TARTARUS_EVENT_MAPPING

EVENT_FORMAT = '@llHHI'
EVENT_SIZE = struct.calcsize(EVENT_FORMAT)


def loop_on_event(event_file, mapping):
    with open(event_file, 'rb') as open_file:

        while True:
            payload = open_file.read(EVENT_SIZE)

            ev_type, code, value = struct.unpack(EVENT_FORMAT, payload)[2:]

            if (ev_type == code == 0) or ev_type == 4:
                continue

            if ev_type == 1:
                if value == 0:
                    value = 'UP'
                elif value == 1:
                    value = 'DOWN'
                else:
                    value = 'REPEAT'

                code = mapping.get(code, code)

                print("Type: EV_KEY, Code: {0}, Value: {1}".format(code, value))
            else:
                print("Type: {0}, Code: {1}, Value: {2}".format(ev_type, code, value))


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('event_file', metavar='EVENT_FILE', type=str, help="Device event file like \"/dev/input/by-id/usb-Razer_Razer_BlackWidow_Chroma-event-kbd\"")
    parser.add_argument('--tartarus', action='store_true', help='Use the tartarus event mapping instead')

    return parser.parse_args()


def run():
    args = parse_args()

    if not os.path.exists(args.event_file):
        print('Event file does not exist', file=sys.stderr)
        sys.exit(1)

    if args.tartarus:
        mapping = TARTARUS_EVENT_MAPPING
    else:
        # mapping = EVENT_MAPPING
        mapping = {}

    print('Starting. Press keys', file=sys.stderr)

    try:
        loop_on_event(args.event_file, mapping)
    except KeyboardInterrupt:
        print("Exiting", file=sys.stderr)


if __name__ == '__main__':
    run()



# SPDX-License-Identifier: GPL-2.0-or-later

"""
Contains functionality specific to keyboard-like devices
"""

from openrazer.client.constants import MACRO_LED_STATIC, MACRO_LED_BLINK
from openrazer.client.devices import RazerDevice as __RazerDevice, BaseDeviceFactory as __BaseDeviceFactory


class RazerKeyboard(__RazerDevice):
    @property
    def game_mode_led(self) -> bool:
        """
        Get game mode LED state

        :return: LED state
        :rtype: bool
        """
        if self.has('game_mode_led'):
            return bool(self._dbus_interfaces['game_mode_led'].getGameMode())
        else:
            raise NotImplementedError()

    @game_mode_led.setter
    def game_mode_led(self, value: bool) -> None:
        """
        Set game mode LED state

        :param value: LED State
        :type value: bool
        """
        if self.has('game_mode_led'):
            if value:
                self._dbus_interfaces['game_mode_led'].setGameMode(True)
            else:
                self._dbus_interfaces['game_mode_led'].setGameMode(False)
        else:
            raise NotImplementedError()

    @property
    def macro_mode_led(self) -> bool:
        """
        Get macro mode LED state

        :return: LED state
        :rtype: bool
        """
        if self.has('macro_mode_led'):
            return bool(self._dbus_interfaces['macro_mode_led'].getMacroMode())
        else:
            raise NotImplementedError()

    @macro_mode_led.setter
    def macro_mode_led(self, value: bool) -> None:
        """
        Set macro mode LED state

        :param value: LED State
        :type value: bool
        """
        if self.has('macro_mode_led'):
            if value:
                self._dbus_interfaces['macro_mode_led'].setMacroMode(True)
            else:
                self._dbus_interfaces['macro_mode_led'].setMacroMode(False)
        else:
            raise NotImplementedError()

    @property
    def macro_mode_led_effect(self) -> int:
        """
        Get macro mode LED effect

        Can get either blinking or static
        :return: Effect ID
        :rtype: int
        """
        if self.has('macro_mode_led_effect'):
            return int(self._dbus_interfaces['macro_mode_led'].getMacroEffect())
        else:
            raise NotImplementedError()

    @macro_mode_led_effect.setter
    def macro_mode_led_effect(self, value: int) -> None:
        """
        Set macro mode LED effect

        Can set to either MACRO_LED_STATIC or MACRO_LED_BLINK
        :param value: Effect ID
        :type value: int
        """
        if self.has('macro_mode_led_effect'):
            if value in (MACRO_LED_STATIC, MACRO_LED_BLINK):
                self._dbus_interfaces['macro_mode_led'].setMacroEffect(value)
            else:
                raise ValueError(f"Unexpected MACRO_LED_* value passed {value}")
        else:
            raise NotImplementedError()

    @property
    def keyswitch_optimization(self) -> bool:
        """
        Get keyswitch optimization state

        :return: Keyswitch optimization state
        :rtype: bool
        """
        if self.has('keyswitch_optimization'):
            return bool(self._dbus_interfaces['keyswitch_optimization'].getKeyswitchOptimization())
        else:
            raise NotImplementedError()

    @keyswitch_optimization.setter
    def keyswitch_optimization(self, value: bool) -> None:
        """
        Set keyswitch optimization state

        :param value: Keyswitch optimization state
        :type value: bool
        """
        if self.has('keyswitch_optimization'):
            self._dbus_interfaces['keyswitch_optimization'].setKeyswitchOptimization(value)
        else:
            raise NotImplementedError()

    @property
    def profile_led_red(self) -> bool:
        """
         Get red profile LED state

         :return: Red profile LED state
         :rtype: bool
         """
        if self.has('lighting_profile_led_red'):
            return bool(self._dbus_interfaces['profile_led'].getRedLED())
        else:
            raise NotImplementedError()

    @profile_led_red.setter
    def profile_led_red(self, enable: bool) -> None:
        """
        Set red profile LED state

        :param enable: Status of red profile LED
        :type enable: bool
        """
        if self.has('lighting_profile_led_red'):
            self._dbus_interfaces['profile_led'].setRedLED(enable)
        else:
            raise NotImplementedError()

    @property
    def profile_led_green(self) -> bool:
        """
        Get green profile LED state

        :return: Green profile LED state
        :rtype: bool
        """
        if self.has('lighting_profile_led_green'):
            return bool(self._dbus_interfaces['profile_led'].getGreenLED())
        else:
            raise NotImplementedError()

    @profile_led_green.setter
    def profile_led_green(self, enable: bool) -> None:
        """
        Set green profile LED state

        :param enable: Status of green profile LED
        :type enable: bool
        """
        if self.has('lighting_profile_led_green'):
            self._dbus_interfaces['profile_led'].setGreenLED(enable)
        else:
            raise NotImplementedError()

    @property
    def profile_led_blue(self) -> bool:
        """
        Get blue profile LED state

        :return: Blue profile LED state
        :rtype: bool
        """
        if self.has('lighting_profile_led_blue'):
            return bool(self._dbus_interfaces['profile_led'].getBlueLED())
        else:
            raise NotImplementedError()

    @profile_led_blue.setter
    def profile_led_blue(self, enable: bool) -> None:
        """
        Set blue profile LED state

        :param enable: Status of blue profile LED
        :type enable: bool
        """
        if self.has('lighting_profile_led_blue'):
            self._dbus_interfaces['profile_led'].setBlueLED(enable)
        else:
            raise NotImplementedError()



# SPDX-License-Identifier: GPL-2.0-or-later

"""
Service Object for DBus
"""

# Disable some pylint stuff
# pylint: disable=no-member

import types
import dbus
import dbus.service


def copy_func(function_reference, name=None):
    """
    Copy function

    :param function_reference: Function
    :type function_reference: func

    :param name: Name of function
    :type name: str

    :return: Copy of function
    :rtype: func
    """
    if hasattr(function_reference, 'code'):
        return types.FunctionType(function_reference.code, function_reference.globals, name or function_reference.func_name, function_reference.defaults, function_reference.closure)
    else:
        return types.FunctionType(function_reference.__code__, function_reference.__globals__, name or function_reference.func_name, function_reference.__defaults__, function_reference.__closure__)


class DBusService(dbus.service.Object):
    """
    DBus Service object

    Allows for dynamic method adding
    """
    BUS_NAME = 'org.razer'

    def __init__(self, object_path):
        """
        Init the object

        :param object_path: DBus Object name
        :type object_path: str
        """
        # We could pass (bus, object_path) here, but we rather register the object manually.
        super().__init__()

        bus = dbus.SessionBus()

        # the constructor of BusName registers the bus, the returned object is not used but must be kept
        self.bus_name_obj = dbus.service.BusName(self.BUS_NAME, bus)

        self.add_to_connection(bus, object_path)

    def add_dbus_method(self, interface_name, function_name, function, in_signature=None, out_signature=None, byte_arrays=False):
        """
        Add method to DBus Object

        :param interface_name: DBus interface name
        :type interface_name: str

        :param function_name: DBus function name
        :type function_name: str

        :param function: Function reference
        :type function: object

        :param in_signature: DBus function signature
        :type in_signature: str

        :param out_signature: DBus function signature
        :type out_signature: str

        :param byte_arrays: Is byte array
        :type byte_arrays: bool
        """

        # Get class key for use in the DBus introspection table
        class_key = [key for key in self._dbus_class_table.keys() if key.endswith(self.__class__.__name__)][0]

        # Create a copy of the function so that if its used multiple times it won't affect other instances if the names changed
        function_deepcopy = copy_func(function, function_name)
        func = dbus.service.method(interface_name, in_signature=in_signature, out_signature=out_signature, byte_arrays=byte_arrays)(function_deepcopy)

        # Add method to DBus tables
        try:
            self._dbus_class_table[class_key][interface_name][function_name] = func
        except KeyError:
            self._dbus_class_table[class_key][interface_name] = {function_name: func}

        # Add method to class as DBus expects it to be there.
        setattr(self.__class__, function_name, func)

    def del_dbus_method(self, interface_name, function_name):
        """
        Remove method from DBus Object

        :param interface_name: DBus interface name
        :type interface_name: str

        :param function_name: DBus function name
        :type function_name: str
        """

        # Get class key for use in the DBus introspection table
        class_key = [key for key in self._dbus_class_table.keys() if key.endswith(self.__class__.__name__)][0]

        # Remove method from DBus tables
        # Remove method from class
        try:
            del self._dbus_class_table[class_key][interface_name][function_name]
            delattr(DBusService, function_name)

        except (KeyError, AttributeError):
            pass



# SPDX-License-Identifier: GPL-2.0-or-later

"""
A class that writes persistence data to disk when device state is updated.

Due to the architecture of the daemon, it's not as simple to know when
something has changed, so for now, an interval will periodically check for
changes in memory and write them to disk. This also avoids excessive
writes when lots of variables change at once.

This is essential because many desktop environments actually kill off
the daemon upon logout/shutdown, thereby persistence isn't retained across
sessions.

A known issue is that this doesn't monitor DPI changes via hardware buttons,
so this won't be persisted until the state is updated via the API.
"""
import time


class PersistenceAutoSave(object):
    def __init__(self, persistence, persistence_file, persistence_status, logger, interval, persistence_save_fn):
        self.persistence = persistence
        self.persistence_file = persistence_file
        self.persistence_status = persistence_status
        self.persistence_save_fn = persistence_save_fn
        self.logger = logger
        self.interval = interval

    def watch(self):
        # Run indefinitely until process is terminated
        while True:
            time.sleep(self.interval)

            if self.persistence_status["changed"]:
                self.logger.debug("State recently changed, writing to disk")
                self.persistence_status["changed"] = False
                self.persistence_save_fn(self.persistence_file)



#!/usr/bin/python3
# SPDX-License-Identifier: GPL-2.0-or-later

from setuptools import setup, find_packages

setup(
    name="openrazer",
    version="3.12.2",
    packages=find_packages(exclude=["tests", "openrazer._fake_driver"]),
    install_requires=[
        "dbus-python >= 1.2.0",
        "numpy >= 1.11.0",
        "openrazer_daemon == 3.12.2",
    ],
    author="OpenRazer contributors",
    description="Library for interacting with the OpenRazer daemon.",
    license="GPLv2+",
    url="https://openrazer.github.io/",
    project_urls={
        "Bug Tracker": "https://github.com/openrazer/openrazer/issues",
        "Source Code": "https://github.com/openrazer/openrazer/tree/master/pylib",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)",
        "Operating System :: POSIX :: Linux",
    ],
)



#!/usr/bin/python3
import argparse
import glob
import os
import sys


def parse_args():
    parser = argparse.ArgumentParser(description="Get battery level")
    parser.add_argument('-d', '--device', type=str, help="Device string like \"0003:1532:0045.000C\"")

    args = parser.parse_args()
    return args


def run():
    args = parse_args()

    if args.device is None:
        mouse_dirs = glob.glob(os.path.join('/sys/bus/hid/drivers/razermouse/', "*:*:*.*"))

        if len(mouse_dirs) > 1:
            print("Multiple mouse directories found. Rerun with -d", file=sys.stderr)
            sys.exit(1)
        if len(mouse_dirs) < 1:
            print("No mouse directories found. Make sure the driver is binded", file=sys.stderr)
            sys.exit(1)

        mouse_dir = mouse_dirs[0]
    else:
        mouse_dir = os.path.join('/sys/bus/hid/drivers/razermouse/', args.device)

    if not os.path.isdir(mouse_dir):
        print("Multiple mouse directories found. Rerun with -d", file=sys.stderr)
        sys.exit(1)

    battery_filepath = os.path.join(mouse_dir, "get_battery")
    with open(battery_filepath, 'r') as battery_file:
        # Battery percentage, before being scaled to 0-100
        try:
            battery_percentage_ns = int(battery_file.read().strip())
            scaled_percentage = (100 / 255) * battery_percentage_ns

            print("{0:.1f}%".format(scaled_percentage))

        except ValueError as ex:
            print("Failed to get battery percentage.\n{0}".format(ex), file=sys.stderr)
            sys.exit(1)


if __name__ == '__main__':
    run()



#!/usr/bin/python3
"""
This script takes in CSV files generated from wireshark's "Export Packet Dissections > as CSV" option.

And prints out nice output.
"""
import argparse
import csv

FIELDS = {
    'status': 0,
    'transaction_id': 1,
    'packet_num': (2, 4),
    'proto_type': 4,
    'data_size': 5,
    'class': 6,
    'command': 7,
    'params': (8, 24)
}


def expand_payload(payload):
    """
    Converts a payload string to a nice format

    :param payload: String of hex digits
    :type payload: str

    :return: Dictionary
    :rtype dict
    """
    chunks = [payload[i:i + 2] for i in range(0, len(payload), 2)]

    result = {}

    for header, num_range in FIELDS.items():
        if isinstance(num_range, tuple):
            result[header] = ''.join(chunks[num_range[0]:num_range[1]])
        else:
            result[header] = chunks[num_range]

    return result


def parse_args():
    """
    Parses command line arguments

    :return: Argparse arguments object
    """
    parser = argparse.ArgumentParser(description="Extracts info from CSV")
    parser.add_argument("file", metavar='FILE', type=str, help="CSV file")

    return parser.parse_args()


def run():
    """
    Main function
    """
    args = parse_args()

    data = []

    with open(args.file, 'r', newline='') as csv_file:
        csv_object = csv.DictReader(csv_file, delimiter=',', quotechar='"')
        for line in csv_object:
            if line['Leftover Capture Data'] == '':
                continue

            # Now we have data
            data.append(expand_payload(line['Leftover Capture Data']))

    format_string = "{0:<6}  {1:<2}  {2:<10}  {3:<4}  {4:<5} {5:<7}  {6}"
    print(format_string.format('Status', 'ID', 'Packet Num', 'Size', 'Class', 'Command', 'Params'))
    for frame in data:
        print(format_string.format(frame['status'], frame['transaction_id'], frame['packet_num'],
                                   frame['data_size'], frame['class'], frame['command'], frame['params']))
    print("")


if __name__ == '__main__':
    run()



#!/usr/bin/python3
# SPDX-License-Identifier: GPL-2.0-or-later

from setuptools import setup, find_packages

setup(
    name="openrazer_daemon",
    version="3.12.2",
    packages=find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests"]),
    install_requires=[
        "daemonize >= 2.4.7",
        "dbus-python >= 1.2.0",
        "PyGObject >= 3.20.0",
        "pyudev >= 0.16.1",
        "setproctitle >= 1.1.8",
    ],
)



# SPDX-License-Identifier: GPL-2.0-or-later

from openrazer.client import constants as constants
from openrazer.client.device_manager import DeviceManager as DeviceManager, DaemonNotFound as DaemonNotFound, __version__ as __version__



#!/usr/bin/python3
import argparse
import atexit
import cmd
import os
import shutil
import sys
import tempfile
import time

PYLIB = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'pylib')
sys.path.insert(1, PYLIB)

import openrazer._fake_driver as fake_driver


class FakeDevicePrompt(cmd.Cmd):

    def __init__(self, device_map, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._device_map = device_map
        self._current_device = None

        self._ep = {}
        self._read = []
        self._write = []

        # If only 1 device, auto use that
        if len(self._device_map) == 1:
            self._change_device(list(self._device_map.keys())[0])
        else:
            self._change_device(None)

    def _change_device(self, device_name=None):
        if device_name is not None:
            self._current_device = device_name
            self.prompt = self._current_device + "> "

            for endpoint, details in self._device_map[self._current_device].endpoints.items():
                self._ep[endpoint] = details[2]

            self._read = [endpoint for endpoint, perm in self._ep.items()]
            self._write = [endpoint for endpoint, perm in self._ep.items() if perm in ('w', 'rw')]
        else:
            self._current_device = None
            self.prompt = "> "

    def do_dev(self, arg):
        """
        Change current device
        """
        if arg in self._device_map:
            if arg is None or len(arg) == 0:
                print('Need to specify a device name. One of: {0}'.format(','.join(self._device_map.keys())))
            else:
                self._change_device(arg)
        else:
            print('Invalid device name: {0}'.format(arg))

    def complete_dev(self, text, line, begidx, endidx):
        if not text:
            completions = list(self._device_map.keys())
        else:
            completions = [item for item in list(self._device_map.keys()) if item.startswith(text)]

        return completions

    def do_list(self, arg):
        """List available device files"""
        if self._current_device is not None:
            print('Device files')
            print('------------')
            for endpoint, permission in self._ep.items():
                if permission in ('r', 'rw'):
                    print("  {0:-<2}-  {1}".format(permission, endpoint))
                else:
                    print("  {0:->2}-  {1}".format(permission, endpoint))

            print()
            print('Event files')
            print('-----------')
            for event_id, event_value in sorted(self._device_map[self._current_device].events.items(), key=lambda x: x[0]):
                print("  {0: >2}   {1}".format(event_id, event_value[0]))
        else:
            print('Devices')
            print('-------')
            for device in list(self._device_map.keys()):
                print('  {0}'.format(device))

    def do_ls(self, arg):
        """List available device files"""
        self.do_list(arg)

    def do_read(self, arg, binary=False):
        """Read ASCII from given device file"""
        if self._current_device is not None:
            if arg in self._ep:
                result = self._device_map[self._current_device].get(arg, binary=binary)

                print(result)
            elif arg in self._ep:
                print('Device endpoint not readable')
            else:
                print("Device endpoint not found")

    def do_binary_read(self, arg):
        """Read binary from given device file"""
        self.do_read(arg, binary=True)

    def complete_read(self, text, line, begidx, endidx):
        if not text:
            completions = self._read
        else:
            completions = [item for item in self._read if item.startswith(text)]

        return completions

    complete_binary_read = complete_read

    def do_write(self, arg):
        """Write ASCII to device file. DEVICE_FILE DATA"""
        if self._current_device is not None:
            try:
                device_file, data = arg.split(' ', 1)

                if device_file in self._ep:
                    if len(data) > 0:
                        self._device_map[self._current_device].set(device_file, data)

                        print("{0}: {1}".format(device_file, self._device_map[self._current_device].get(device_file)))
                else:
                    print("Device endpoint not found")

            except ValueError:
                print("Must specify a device endpoint then a space then data to write")

    def complete_write(self, text, line, begidx, endidx):
        if not text:
            completions = self._write
        else:
            completions = [item for item in self._write if item.startswith(text)]

        return completions

    def do_event(self, arg):
        """Emit an event, format: EVENT_ID KEY_ID STATE

        Where state in 'up' 'down' and 'repeat'
        """
        if self._current_device is not None:
            try:
                event_file, key_id, value = arg.split(' ')
            except ValueError:
                print("Usage: event event_file key_id value")
                return

            if event_file not in self._device_map[self._current_device].events:
                print("Event ID {0} is invalid".format(event_file))
            else:
                try:
                    bytes_written = self._device_map[self._current_device].emit_kb_event(event_file, int(key_id), value)
                    print("Wrote {0} bytes to {1}".format(bytes_written, self._device_map[self._current_device].events[event_file][0]))
                except ValueError as err:
                    print("Caught exception: {0}".format(err))

    def do_exit(self, arg):
        """Exit"""
        if self._current_device is not None:
            self._change_device(None)
            return False
        else:
            return True

    def do_EOF(self, arg):
        """Press Ctrl+D to exit"""
        self.do_exit(arg)


def create_envionment(device_name, destination):
    os.makedirs(destination, exist_ok=True)

    try:
        fake_device = fake_driver.FakeDevice(device_name, tmp_dir=destination)
        return fake_device
    except ValueError:
        print('Device {0}.cfg not found'.format(device_name))


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument('device', metavar='DEVICE', nargs='*', help='Device config name')
    parser.add_argument('--dest', metavar='DESTDIR', required=False, default=None, help='Directory to create driver files in. If omitted then a tmp directory is used')
    parser.add_argument('--all', action='store_true', help='Create all possible fake devices')
    parser.add_argument('--non-interactive', dest='interactive', action='store_false', help='Dont display prompt, just hang until killed')
    parser.add_argument('--clear-dest', action='store_true', help='Clear the destination folder if it exists before starting')

    return parser.parse_args()


def run():
    args = parse_args()

    if args.dest is None:
        destination = tempfile.mkdtemp(prefix='tmp_', suffix='_fakerazer')
    else:
        destination = args.dest

        if args.clear_dest and os.path.exists(destination):
            shutil.rmtree(destination, ignore_errors=True)

    if args.all:
        devices = fake_driver.SPECS
    else:
        devices = args.device

    device_map = {}
    for device in devices:
        # Device name: FakeDriver
        fake_device = create_envionment(device, destination)
        if fake_device is not None:
            device_map[device] = fake_device

    if len(device_map) == 0:
        print("ERROR: No valid devices passed to script, you either need to pass devices as arguments or use '--all'")
        sys.exit(1)

    # Register cleanup
    if args.dest is None:
        atexit.register(lambda: shutil.rmtree(destination, ignore_errors=True))
    else:
        for device in device_map.values():
            # device = FakeDriver
            atexit.register(device.close)

    print("Device test directory: {0}".format(destination))

    try:
        if not args.interactive:
            print("Sleeping forever, use Ctrl-C to exit...")
            while True:
                time.sleep(99999999)
        else:
            FakeDevicePrompt(device_map).cmdloop()
    except KeyboardInterrupt:
        pass


if __name__ == '__main__':
    run()



import colorsys
import random

from openrazer.client import DeviceManager
from openrazer.client import constants as razer_constants

# Create a DeviceManager. This is used to get specific devices
device_manager = DeviceManager()


print("Found {} Razer devices".format(len(device_manager.devices)))
print()

# Disable daemon effect syncing.
# Without this, the daemon will try to set the lighting effect to every device.
device_manager.sync_effects = False

# List of effect I've chosen to make an example for
effects = [
    'breath_random',
    'breath_single',
    'breath_dual',
    'breath_triple',
    'reactive',
    'spectrum',
    'static',
    'wave',
]

# Helper function to generate interesting colors


def random_color():
    rgb = colorsys.hsv_to_rgb(random.uniform(0, 1), random.uniform(0.5, 1), 1)
    return tuple(map(lambda x: int(256 * x), rgb))


# Iterate over each device and set a random effect that it supports.
for device in device_manager.devices:
    # Check which effect this device supports.
    device_effects = [effect for effect in effects if device.fx.has(effect)]
    # print("{} supports {}".format(device.name, device_effects))

    if len(device_effects) == 0:
        print("Device {} doesn't support any of the effects".format(device.name))
        continue
    effect = random.choice(device_effects)
    print("Setting {} to effect {}".format(device.name, effect))

    # Ad an example for each effect
    if effect == 'breath_random':
        device.fx.breath_random()

    elif effect == 'breath_single':
        color = random_color()
        device.fx.breath_single(color[0], color[1], color[2])

    elif effect == 'breath_dual':
        color = random_color()
        color2 = random_color()
        device.fx.breath_dual(color[0], color[1], color[2],
                              color2[0], color2[1], color2[2])

    elif effect == 'breath_triple':
        color = random_color()
        color2 = random_color()
        color3 = random_color()
        device.fx.breath_triple(color[0], color[1], color[2],
                                color2[0], color2[1], color2[2],
                                color3[0], color3[1], color3[2])

    elif effect == 'reactive':
        color = random_color()
        times = [razer_constants.REACTIVE_500MS, razer_constants.REACTIVE_1000MS,
                 razer_constants.REACTIVE_1500MS, razer_constants.REACTIVE_2000MS]
        device.fx.reactive(color[0], color[1], color[2], random.choice(times))

    elif effect == 'spectrum':
        device.fx.spectrum()

    elif effect == 'static':
        color = random_color()
        device.fx.static(*color)

    elif effect == 'wave':
        directions = [razer_constants.WAVE_LEFT, razer_constants.WAVE_RIGHT]
        device.fx.wave(random.choice(directions))



# SPDX-License-Identifier: GPL-2.0-or-later

"""
Class to manage syncing of effects
"""
import inspect
import logging


class EffectSync(object):
    """
    Class which deals with receiving effect events from other devices
    """

    def __init__(self, parent, device_number):
        self._logger = logging.getLogger('razer.device{0}.effect_sync'.format(device_number))
        self._parent = parent
        self._parent.register_observer(self)

    def __del__(self):
        self.close()

    def close(self):
        self._parent.remove_observer(self)

    def notify(self, msg):
        """
        Receive notificatons from the device (we only care about effects)

        :param msg: Notification
        :type msg: tuple
        """
        if not isinstance(msg, tuple):
            self._logger.warning("Got msg that was not a tuple")
        elif msg[0] == 'effect':
            # We have a message directed at us
            # MSG format
            #  0         1       2             3
            # ('effect', Device, 'effectName', 'effectparams'...)
            # Device is the device the msg originated from (could be parent device)
            if msg[1] is not self._parent:
                # Msg from another device
                self.run_effect(msg[2], *msg[3:])

    def run_effect(self, effect_name, *args):
        """
        Run the specified effect with the given arguments

        :param effect_name: Name of the effect
        :type effect_name: str

        :param args: Arguments for the specified effect
        :type args: list
        """
        # Disable notifications
        self._parent.disable_notify = True

        try:
            # Does parent have method
            effect_func = getattr(self._parent, effect_name, None)
            if effect_func is not None:
                # We have method, does it have the correct num arguments
                actual_args = self.get_num_arguments(effect_func)
                if actual_args == len(args):
                    # method should be same
                    effect_func(*args)
                else:
                    # Method same but wrong args, try alternatives
                    if effect_name == 'setStatic':
                        # Could be static from chroma to non chroma
                        if actual_args == 0:
                            # Chroma -> BW
                            effect_func()
                        else:
                            # BW -> Chroma
                            effect_func(0x0, 0xFF, 0x00)  # Green

            else:
                # setNone sets active to false and needs to be re-enabled for effects to show - maybe a bit inefficient
                if not effect_name == 'setNone':
                    effect_func = getattr(self._parent, 'setScrollActive', None)
                    if effect_func is not None:
                        effect_func(True)
                    effect_func = getattr(self._parent, 'setLogoActive', None)
                    if effect_func is not None:
                        effect_func(True)
                    effect_func = getattr(self._parent, 'setLeftActive', None)
                    if effect_func is not None:
                        effect_func(True)
                    effect_func = getattr(self._parent, 'setRightActive', None)
                    if effect_func is not None:
                        effect_func(True)
                    effect_func = getattr(self._parent, 'setBacklightActive', None)
                    if effect_func is not None:
                        effect_func(True)
                else:
                    effect_func = getattr(self._parent, 'setScrollNone', None)
                    if effect_func is not None:
                        effect_func()
                    effect_func = getattr(self._parent, 'setLogoNone', None)
                    if effect_func is not None:
                        effect_func()
                    effect_func = getattr(self._parent, 'setLeftNone', None)
                    if effect_func is not None:
                        effect_func()
                    effect_func = getattr(self._parent, 'setRightNone', None)
                    if effect_func is not None:
                        effect_func()
                    effect_func = getattr(self._parent, 'setBacklightNone', None)
                    if effect_func is not None:
                        effect_func()

                # The target device doesn't have these methods, use similar ones

                if effect_name == 'setPulsate':
                    # setPulsate doesn't provide a color but we need one, take green.
                    pargs = (0x00, 0xFF, 0x00)  # Green
                    effect_func = getattr(self._parent, 'setBreathSingle', None)
                    if effect_func is not None:
                        effect_func(*pargs)
                    effect_func = getattr(self._parent, 'setScrollBreathSingle', None)
                    if effect_func is not None:
                        effect_func(*pargs)
                    effect_func = getattr(self._parent, 'setLogoBreathSingle', None)
                    if effect_func is not None:
                        effect_func(*pargs)
                    effect_func = getattr(self._parent, 'setLeftBreathSingle', None)
                    if effect_func is not None:
                        effect_func(*pargs)
                    effect_func = getattr(self._parent, 'setRightBreathSingle', None)
                    if effect_func is not None:
                        effect_func(*pargs)
                    effect_func = getattr(self._parent, 'setScrollPulsate', None)
                    if effect_func is not None:
                        effect_func(*pargs)
                    effect_func = getattr(self._parent, 'setLogoPulsate', None)
                    if effect_func is not None:
                        effect_func(*pargs)
                    effect_func = getattr(self._parent, 'setBacklightPulsate', None)
                    if effect_func is not None:
                        effect_func(*pargs)

                elif effect_name == 'setSpectrum':
                    effect_func = getattr(self._parent, 'setScrollSpectrum', None)
                    if effect_func is not None:
                        effect_func()
                    effect_func = getattr(self._parent, 'setLogoSpectrum', None)
                    if effect_func is not None:
                        effect_func()
                    effect_func = getattr(self._parent, 'setLeftSpectrum', None)
                    if effect_func is not None:
                        effect_func()
                    effect_func = getattr(self._parent, 'setRightSpectrum', None)
                    if effect_func is not None:
                        effect_func()
                    effect_func = getattr(self._parent, 'setBacklightSpectrum', None)
                    if effect_func is not None:
                        effect_func()

                elif effect_name == 'setStatic':
                    effect_func = getattr(self._parent, 'setScrollStatic', None)
                    if effect_func is not None:
                        effect_func(*args)
                    effect_func = getattr(self._parent, 'setLogoStatic', None)
                    if effect_func is not None:
                        effect_func(*args)
                    effect_func = getattr(self._parent, 'setLeftStatic', None)
                    if effect_func is not None:
                        effect_func(*args)
                    effect_func = getattr(self._parent, 'setRightStatic', None)
                    if effect_func is not None:
                        effect_func(*args)
                    effect_func = getattr(self._parent, 'setBacklightStatic', None)
                    if effect_func is not None:
                        effect_func(*args)

                elif effect_name == 'setWave':
                    effect_func = getattr(self._parent, 'setScrollWave', None)
                    if effect_func is not None:
                        effect_func(*args)
                    effect_func = getattr(self._parent, 'setLogoWave', None)
                    if effect_func is not None:
                        effect_func(*args)
                    effect_func = getattr(self._parent, 'setLeftWave', None)
                    if effect_func is not None:
                        effect_func(*args)
                    effect_func = getattr(self._parent, 'setRightWave', None)
                    if effect_func is not None:
                        effect_func(*args)
                    effect_func = getattr(self._parent, 'setBacklightWave', None)
                    if effect_func is not None:
                        effect_func(*args)

                elif effect_name == 'setReactive':
                    effect_func = getattr(self._parent, 'setScrollReactive', None)
                    if effect_func is not None:
                        effect_func(*args)
                    effect_func = getattr(self._parent, 'setLogoReactive', None)
                    if effect_func is not None:
                        effect_func(*args)
                    effect_func = getattr(self._parent, 'setLeftReactive', None)
                    if effect_func is not None:
                        effect_func(*args)
                    effect_func = getattr(self._parent, 'setRightReactive', None)
                    if effect_func is not None:
                        effect_func(*args)
                    effect_func = getattr(self._parent, 'setBacklightReactive', None)
                    if effect_func is not None:
                        effect_func(*args)

                elif effect_name in ('setBreathRandom', 'setBreathSingle', 'setBreathDual', 'setBreathTriple'):
                    if effect_name in ('setBreathRandom', 'setBreathSingle'):
                        pargs = (0x00, 0xFF, 0x00)  # Green
                    else:
                        pargs = args[0:3]  # limit args to first 3, as setBreathDual gives 6 args and setBreathTriple gives 9 args

                    if effect_name == 'setBreathRandom':
                        effect_func = getattr(self._parent, 'setPulsate', None)
                        if effect_func is not None:
                            # setPulsate doesn't take any argument
                            effect_func()
                        effect_func = getattr(self._parent, 'setScrollPulsate', None)
                        if effect_func is not None:
                            effect_func(*pargs)
                        effect_func = getattr(self._parent, 'setLogoPulsate', None)
                        if effect_func is not None:
                            effect_func(*pargs)
                        effect_func = getattr(self._parent, 'setBacklightPulsate', None)
                        if effect_func is not None:
                            effect_func(*pargs)

                        effect_func = getattr(self._parent, 'setScrollBreathRandom', None)
                        if effect_func is not None:
                            effect_func(*args)
                        effect_func = getattr(self._parent, 'setLogoBreathRandom', None)
                        if effect_func is not None:
                            effect_func(*args)
                        effect_func = getattr(self._parent, 'setLeftBreathRandom', None)
                        if effect_func is not None:
                            effect_func(*args)
                        effect_func = getattr(self._parent, 'setRightBreathRandom', None)
                        if effect_func is not None:
                            effect_func(*args)
                        effect_func = getattr(self._parent, 'setBacklightBreathRandom', None)
                        if effect_func is not None:
                            effect_func(*args)

                    if effect_name == 'setBreathSingle':
                        if effect_func is not None:
                            effect_func(*pargs)
                        effect_func = getattr(self._parent, 'setScrollPulsate', None)
                        if effect_func is not None:
                            effect_func(*pargs)
                        effect_func = getattr(self._parent, 'setLogoPulsate', None)
                        if effect_func is not None:
                            effect_func(*pargs)
                        effect_func = getattr(self._parent, 'setBacklightPulsate', None)
                        if effect_func is not None:
                            effect_func(*pargs)

                        effect_func = getattr(self._parent, 'setScrollBreathSingle', None)
                        if effect_func is not None:
                            effect_func(*args)
                        effect_func = getattr(self._parent, 'setLogoBreathSingle', None)
                        if effect_func is not None:
                            effect_func(*args)
                        effect_func = getattr(self._parent, 'setLeftBreathSingle', None)
                        if effect_func is not None:
                            effect_func(*args)
                        effect_func = getattr(self._parent, 'setRightBreathSingle', None)
                        if effect_func is not None:
                            effect_func(*args)
                        effect_func = getattr(self._parent, 'setBacklightBreathSingle', None)
                        if effect_func is not None:
                            effect_func(*args)

                    if effect_name == 'setBreathDual':
                        effect_func = getattr(self._parent, 'setScrollBreathDual', None)
                        if effect_func is not None:
                            effect_func(*args)
                        effect_func = getattr(self._parent, 'setLogoBreathDual', None)
                        if effect_func is not None:
                            effect_func(*args)
                        effect_func = getattr(self._parent, 'setLeftBreathDual', None)
                        if effect_func is not None:
                            effect_func(*args)
                        effect_func = getattr(self._parent, 'setRightBreathDual', None)
                        if effect_func is not None:
                            effect_func(*args)
                        effect_func = getattr(self._parent, 'setBacklightBreathDual', None)
                        if effect_func is not None:
                            effect_func(*args)

                elif effect_name == 'setBrightness':
                    effect_func = getattr(self._parent, 'setScrollBrightness', None)
                    if effect_func is not None:
                        effect_func(*args)
                    effect_func = getattr(self._parent, 'setLogoBrightness', None)
                    if effect_func is not None:
                        effect_func(*args)
                    effect_func = getattr(self._parent, 'setLeftBrightness', None)
                    if effect_func is not None:
                        effect_func(*args)
                    effect_func = getattr(self._parent, 'setRightBrightness', None)
                    if effect_func is not None:
                        effect_func(*args)
                    effect_func = getattr(self._parent, 'setBacklightBrightness', None)
                    if effect_func is not None:
                        effect_func(*args)

        except Exception as err:
            self._logger.exception("Caught exception trying to sync effects.", exc_info=err)

        # Re-enable notifications
        self._parent.disable_notify = False

    @staticmethod
    def get_num_arguments(func):
        """
        Get number of arguments in a function

        :param func: Function
        :type func: callable

        :return: Number of arguments
        :rtype: int
        """
        func_sig = inspect.signature(func)
        return len(func_sig.parameters)



#!/usr/bin/python3
# SPDX-License-Identifier: GPL-2.0-or-later

import argparse
import os
import shutil
import signal
import sys
import time
import logging

from openrazer_daemon.daemon import RazerDaemon, __version__
from subprocess import check_output
from time import sleep
from daemonize import Daemonize

# Basically copied from https://github.com/jleclanche/python-xdg/blob/master/xdg/basedir.py
HOME = os.path.expanduser("~")
XDG_DATA_HOME = os.environ.get("XDG_DATA_HOME", os.path.join(HOME, ".local", "share"))
XDG_CONFIG_HOME = os.environ.get("XDG_CONFIG_HOME", os.path.join(HOME, ".config"))

RAZER_DATA_HOME = os.path.join(XDG_DATA_HOME, "openrazer")
XDG_RUNTIME_DIR = os.environ.get("XDG_RUNTIME_DIR", RAZER_DATA_HOME)
RAZER_CONFIG_HOME = os.path.join(XDG_CONFIG_HOME, "openrazer")
RAZER_RUNTIME_DIR = XDG_RUNTIME_DIR

EXAMPLE_CONF_FILE = '/usr/share/openrazer/razer.conf.example'

CONF_FILE = os.path.join(RAZER_CONFIG_HOME, 'razer.conf')
PERSISTENCE_FILE = os.path.join(RAZER_CONFIG_HOME, 'persistence.conf')
LOG_PATH = os.path.join(RAZER_DATA_HOME, 'logs')

args = None


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument('-v', '--verbose', action='store_true', help='Enable verbose logging')

    parser.add_argument('-F', '--foreground', action='store_true', help='Don\'t fork stay in the foreground')

    parser.add_argument('-r', '--respawn', action='store_true', help='Stop any existing daemon first, if one is running.')
    parser.add_argument('-s', '--stop', action='store_true', help='Gracefully stop the existing daemon.')
    parser.add_argument('--version', action='version', version='%(prog)s {version}'.format(version=__version__))

    parser.add_argument('--as-root', action='store_true', help='Allow the daemon to be started as root')

    parser.add_argument('--config', type=str, help='Location of the config file', default=CONF_FILE)
    parser.add_argument('--persistence', type=str, help='Location to file for storing device persistence data', default=PERSISTENCE_FILE)
    parser.add_argument('--run-dir', type=str, help='Location of the run directory', default=RAZER_RUNTIME_DIR)
    parser.add_argument('--log-dir', type=str, help='Location of the log directory', default=LOG_PATH)

    parser.add_argument('--test-dir', type=str, help='Directory containing test driver structure')

    return parser.parse_args()


def stop_daemon(args):
    pidfile = os.path.join(args.run_dir, 'openrazer-daemon.pid')
    try:
        with open(pidfile) as f:
            pid = int(f.readline().strip())

            # if we have psutil, check that the process name matches the
            # pidfile. Otherwise we might terminate a process that's not
            # ours.
            try:
                import psutil
                try:
                    if psutil.Process(pid).name() != "openrazer-daemon":
                        raise ProcessLookupError()
                except psutil.NoSuchProcess:
                    raise ProcessLookupError()
            except ImportError:
                print("Module psutil is missing, not checking for process name")

            os.kill(pid, signal.SIGTERM)
            pid_exists = True
            delay = 3000
            while delay > 0:
                delay -= 100
                try:
                    time.sleep(0.1)
                    os.kill(pid, 0)
                except ProcessLookupError:
                    print("Process {} stopped".format(pid))
                    pid_exists = False
                    break

            # if we have to kill it, we probably need to remove the
            # pidfile too, otherwise we rely on it to clean up properly
            if pid_exists:
                print("Process {} is hung, sending SIGKILL".format(pid))
                os.kill(pid, signal.SIGKILL)
                os.remove(pidfile)

    except FileNotFoundError:
        print("No pidfile found, assuming openrazer-daemon is not running")
    except ProcessLookupError:
        print("pidfile exists but no process is running. Remove {} and continue".format(pidfile))


def install_example_config_file(config_file):
    """
    Installs the example config file
    """
    if os.path.exists(config_file):
        return

    try:
        os.makedirs(os.path.dirname(config_file), exist_ok=True)
        if os.path.exists(EXAMPLE_CONF_FILE):
            shutil.copy(EXAMPLE_CONF_FILE, config_file)
        else:
            print('Cant find "{0}"'.format(EXAMPLE_CONF_FILE), file=sys.stderr)
    except NotADirectoryError as e:
        print("Failed to create {}".format(e.filename), file=sys.stderr)
        sys.exit(1)


def init_persistence_config(persistence_file):
    """
    Creates a new file for persistence, if it does not exist.
    """
    if os.path.exists(persistence_file):
        return

    try:
        os.makedirs(os.path.dirname(persistence_file), exist_ok=True)
        with open(persistence_file, "w") as f:
            f.writelines("")

    except NotADirectoryError as e:
        print("Failed to create {}".format(e.filename), file=sys.stderr)
        sys.exit(1)


def run_daemon():
    global args
    daemon = RazerDaemon(verbose=args.verbose,
                         log_dir=args.log_dir,
                         console_log=args.foreground,
                         config_file=args.config,
                         persistence_file=args.persistence,
                         test_dir=args.test_dir)
    try:
        daemon.run()
    except KeyboardInterrupt:
        daemon.logger.debug("Exited on user request")
    except Exception as err:
        daemon.logger.exception("Caught exception", exc_info=err)


def run():
    global args

    logger = None
    args = parse_args()

    if args.stop:
        stop_daemon(args)
        sys.exit(0)

    if os.getuid() == 0:
        if args.as_root:
            print("The daemon is being run as root.")
        else:
            print("The daemon should not be run as root. If you have a good reason to do so, use the --as-root flag.")
            sys.exit(1)

    if args.respawn:
        stop_daemon(args)
        time.sleep(3)

    # daemonize logs exceptions to its logger (which defaults to the syslog)
    # and does not make them appear on stdout/stderr. If we're in foreground
    # mode, override that logger with our own.
    if args.foreground:
        logger = logging.getLogger('run-daemon')
        if args.verbose:
            logger.setLevel(logging.DEBUG)

    install_example_config_file(args.config)
    init_persistence_config(args.persistence)

    os.makedirs(args.run_dir, exist_ok=True)
    daemon = Daemonize(app="openrazer-daemon",
                       pid=os.path.join(args.run_dir, "openrazer-daemon.pid"),
                       action=run_daemon,
                       foreground=args.foreground,
                       verbose=args.verbose,
                       chdir=args.run_dir,
                       logger=logger)
    daemon.start()


if __name__ == "__main__":
    run()



# SPDX-License-Identifier: GPL-2.0-or-later

import unittest

import openrazer_daemon.device

DEVICE1_SERIAL = 'XX000000'
DEVICE1_ID = '0000:0000:0000.0000'

DEVICE2_SERIAL = 'XX000001'
DEVICE2_ID = '0000:0000:0000.0001'


class DummyDBusObject(object):
    def __init__(self):
        self.notify_msg = None
        self.parent = None

    def notify(self, msg):
        self.notify_msg = msg

    def register_parent(self, parent):
        self.parent = parent

    def notify_parent(self, msg):
        self.parent.notify_parent(msg)


class DummyParentObject(object):
    def __init__(self):
        self.notify_msg = None
        self.notify_device = None

    def notify(self, device_object, msg):
        self.notify_device = device_object
        self.notify_msg = msg

# TODO move device_object creation to setUp


class DeviceTest(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_device_properties(self):
        dbus_object = DummyDBusObject()
        device_object = openrazer_daemon.device.Device(DEVICE1_ID, DEVICE1_SERIAL, dbus_object)

        self.assertEqual(device_object.device_id, DEVICE1_ID)
        self.assertEqual(device_object.serial, DEVICE1_SERIAL)
        self.assertEqual(device_object.dbus, dbus_object)

    def test_device_register_parent(self):
        dbus_object = DummyDBusObject()
        parent_object = DummyParentObject()

        device_object = openrazer_daemon.device.Device(DEVICE1_ID, DEVICE1_SERIAL, dbus_object)
        device_object.register_parent(parent_object)

        self.assertEqual(device_object._parent, parent_object)

    def test_device_notify_child(self):
        msg = ('test', 1)

        dbus_object = DummyDBusObject()

        device_object = openrazer_daemon.device.Device(DEVICE1_ID, DEVICE1_SERIAL, dbus_object)
        device_object.notify_child(msg)

        self.assertEqual(dbus_object.notify_msg, msg)

    def test_device_notify_parent(self):
        msg = ('test', 1)

        dbus_object = DummyDBusObject()
        parent_object = DummyParentObject()

        device_object = openrazer_daemon.device.Device(DEVICE1_ID, DEVICE1_SERIAL, dbus_object)
        device_object.register_parent(parent_object)

        device_object.notify_parent(msg)

        self.assertEqual(parent_object.notify_msg, msg)
        self.assertEqual(parent_object.notify_device, device_object)


class DeviceCollectionTest(unittest.TestCase):
    def setUp(self):
        self.device_collection = openrazer_daemon.device.DeviceCollection()

    def test_add(self):
        dbus_object = DummyDBusObject()

        self.device_collection.add(DEVICE1_ID, DEVICE1_SERIAL, dbus_object)

        self.assertIn(DEVICE1_ID, self.device_collection._id_map)
        self.assertIn(DEVICE1_SERIAL, self.device_collection._serial_map)

        device_obj_from_id = self.device_collection._id_map[DEVICE1_ID]
        device_obj_from_serial = self.device_collection._serial_map[DEVICE1_SERIAL]

        self.assertIs(device_obj_from_id, device_obj_from_serial)

    def test_get(self):
        dbus_object = DummyDBusObject()

        self.device_collection.add(DEVICE1_ID, DEVICE1_SERIAL, dbus_object)

        device_obj_by_id = self.device_collection[DEVICE1_ID]
        device_obj_by_serial = self.device_collection[DEVICE1_SERIAL]

        self.assertIs(device_obj_by_id, device_obj_by_serial)

    def test_invalid_get(self):
        try:
            device = self.device_collection.get('INVALID')
        except IndexError:
            pass

    def test_contains(self):
        dbus_object = DummyDBusObject()

        self.device_collection.add(DEVICE1_ID, DEVICE1_SERIAL, dbus_object)

        self.assertIn(DEVICE1_ID, self.device_collection)
        self.assertIn(DEVICE1_SERIAL, self.device_collection)

    def test_remove(self):
        dbus_object = DummyDBusObject()

        self.device_collection.add(DEVICE1_ID, DEVICE1_SERIAL, dbus_object)
        self.assertIn(DEVICE1_ID, self.device_collection)

        self.device_collection.remove(DEVICE1_ID)
        self.assertNotIn(DEVICE1_ID, self.device_collection)

        self.device_collection.add(DEVICE1_ID, DEVICE1_SERIAL, dbus_object)
        self.assertIn(DEVICE1_ID, self.device_collection)

        self.device_collection.remove(DEVICE1_SERIAL)
        self.assertNotIn(DEVICE1_SERIAL, self.device_collection)

    def test_items(self):
        dbus_object = DummyDBusObject()

        self.device_collection.add(DEVICE1_ID, DEVICE1_SERIAL, dbus_object)

        device_id, device_obj1 = list(self.device_collection.id_items())[0]
        device_serial, device_obj2 = list(self.device_collection.serial_items())[0]

        self.assertEqual(device_id, DEVICE1_ID)
        self.assertEqual(device_serial, DEVICE1_SERIAL)
        self.assertIs(device_obj1, device_obj2)

    def test_iter(self):
        dbus_object = DummyDBusObject()

        self.device_collection.add(DEVICE1_ID, DEVICE1_SERIAL, dbus_object)

        devices = [self.device_collection.get(DEVICE1_ID)]

        for device in self.device_collection:
            devices.remove(device)

        self.assertEqual(len(devices), 0)

    def test_serials(self):
        dbus_object1 = DummyDBusObject()
        dbus_object2 = DummyDBusObject()

        self.device_collection.add(DEVICE1_ID, DEVICE1_SERIAL, dbus_object1)
        self.device_collection.add(DEVICE2_ID, DEVICE2_SERIAL, dbus_object2)

        serials = self.device_collection.serials()

        self.assertIn(DEVICE1_SERIAL, serials)
        self.assertIn(DEVICE2_SERIAL, serials)

    def test_devices(self):
        dbus_object1 = DummyDBusObject()
        dbus_object2 = DummyDBusObject()

        self.device_collection.add(DEVICE1_ID, DEVICE1_SERIAL, dbus_object1)
        self.device_collection.add(DEVICE2_ID, DEVICE2_SERIAL, dbus_object2)

        device_list = self.device_collection.devices
        available_dbus = [dbus_object1, dbus_object2]

        for device in device_list:
            available_dbus.remove(device.dbus)

        # Ensure both dbus objects have been seen
        self.assertEqual(len(available_dbus), 0)

    def test_cross_device_notify(self):
        dbus_object1 = DummyDBusObject()
        dbus_object2 = DummyDBusObject()
        msg = ('test', 1)

        self.device_collection.add(DEVICE1_ID, DEVICE1_SERIAL, dbus_object1)
        self.device_collection.add(DEVICE2_ID, DEVICE2_SERIAL, dbus_object2)

        self.assertIs(dbus_object1.notify_msg, None)
        self.assertIs(dbus_object2.notify_msg, None)

        dbus_object1.notify_parent(msg)

        # Ensure message gets sent to other devices and not itself
        self.assertIs(dbus_object1.notify_msg, None)
        self.assertIs(dbus_object2.notify_msg, msg)



#!/usr/bin/python3
import argparse
import glob
import os
import time
import random
import sys


def clamp_to_min_max(value, min, max):
    if value > max:
        value = max
    elif value < min:
        value = min
    return value


def clamp_to_u8(value):
    return clamp_to_min_max(value, 0, 255)


def parse_args():
    parser = argparse.ArgumentParser(description="Set a custom colour effect")
    parser.add_argument('-d', '--device', type=str, help="Device string like \"0003:1532:0045.000C\"")

    args = parser.parse_args()
    return args


def run():
    args = parse_args()

    if args.device is None:
        mouse_dirs = glob.glob(os.path.join('/sys/bus/hid/drivers/razermouse/', "*:*:*.*"))

        if len(mouse_dirs) > 1:
            print("Multiple mouse directories found. Rerun with -d", file=sys.stderr)
            sys.exit(1)
        if len(mouse_dirs) < 1:
            print("No mouse directories found. Make sure the driver is binded", file=sys.stderr)
            sys.exit(1)

        mouse_dir = mouse_dirs[0]
    else:
        mouse_dir = os.path.join('/sys/bus/hid/drivers/razermouse/', args.device)

    if not os.path.isdir(mouse_dir):
        print("Multiple mouse directories found. Rerun with -d", file=sys.stderr)
        sys.exit(1)

    # choices = [b'\xff\x00\x00', b'\xff\xff\x00', b'\x00\xff\x00', b'\x00\xff\xff', b'\x00\x00\xff', b'\xff\x00\xff']
    choices = [b'\xff\x00\x00']

    for repeat in range(0, 10):
        payload = b''
        for i in range(0, 15):
            payload += random.choice(choices)

        set_colour_filename = os.path.join(mouse_dir, "set_key_row")
        set_custom_mode_filename = os.path.join(mouse_dir, "mode_custom")
        with open(set_colour_filename, 'wb') as set_colour_file:
            set_colour_file.write(payload)
        with open(set_custom_mode_filename, 'w') as set_custom_mode_file:
            set_custom_mode_file.write("1")
        time.sleep(0.2)

    print("Done")


if __name__ == '__main__':
    run()



import sys

import openrazer.client

device_manager = openrazer.client.DeviceManager()
keyboard = None


def on_off(value, true='On', false='Off'):
    if value:
        return true
    else:
        return false


for device in device_manager.devices:
    if device.type == 'keyboard':
        keyboard = device
        break
else:
    print("Could not find suitable keyboard", file=sys.stderr)
    sys.exit(1)

print("Name: {0}".format(keyboard.name))
print("Type: {0}".format(keyboard.type))
print("Serial: {0}".format(keyboard.serial))
print("Firmware Version: {0}".format(keyboard.firmware_version))
print("------------------------")
print("Brightness: {0}".format(keyboard.brightness))
if keyboard.has("game_mode_led"):
    print("Game Mode LED: {0}".format(on_off(keyboard.game_mode_led)))
if keyboard.has("macro_mode_led"):
    macro_led = on_off(keyboard.macro_mode_led)
    macro_led_state = on_off(keyboard.macro_mode_led_effect == openrazer.client.constants.MACRO_LED_STATIC, true='Static', false='Blinking')
    print("Macro Mode LED: {0} ({1})".format(macro_led, macro_led_state))
print("------------------------")
print("Capabilities:")
for key, value in sorted(keyboard.capabilities.items(), key=lambda x: x[0]):
    print("  {0}: {1}".format(key, value))



#!/usr/bin/python3
# Copyright 2025 Urufusan.
# SPDX-License-Identifier: GPL-2.0-or-later

# * This example is a daemon that displays the current battery percentage of your mouse with the status LED on a Razer Mouse Dock Chroma
# * https://www.razer.com/gaming-mice-accessories/razer-mouse-dock-chroma

import os
import time
from typing import Sequence

import openrazer.client
from openrazer.client.devices import RazerDevice


def hex_to_rgb(hex_string: str) -> tuple[int, int, int]:
    hex_string = hex_string.lstrip("#")
    _r_hex = hex_string[0:2]
    _g_hex = hex_string[2:4]
    _b_hex = hex_string[4:6]
    return int(_r_hex, 16), int(_g_hex, 16), int(_b_hex, 16)


# * Supported mice for the V1 chroma dock, can be adapted to support V2
CHROMA_DOCK_MICE_NAMES = ["Razer Naga Pro", "Razer DeathAdder V2 Pro", "Razer Basilisk Ultimate", "Razer Viper Ultimate"]

# * Battery % color gradient (0 - red, 100 - green)
GRADIENT = (
    {"color": "#220000", "position": 0},
    {"color": "#ff0000", "position": 20},
    {"color": "#ffa600", "position": 50},
    {"color": "#00ff00", "position": 100},
)

GRADIENT_RGB = [{"color": hex_to_rgb(_grad_clr["color"]), "position": _grad_clr["position"]} for _grad_clr in GRADIENT]

POLL_DELAY_CONST = int(os.environ.get("RZR_SLEEPFOR", 360))
QUICK_RETRIES_AMOUT = 6  # Does a shallow check, doesn't reinit dbus
FULL_RETRIES_AMOUT = 2  # Does a 'full' refresh (devices that weren't there before will now be shown); reinits DeviceManager


def pick_gradient_color(value: int | float, gradient_map: Sequence[dict]) -> tuple[int, int, int]:
    """## Generates color based on the input value

    Define your gradient colors and positions with either floats or ints:
    ```
    gradient = [
       {'color': '#03ee80', 'position': 0},
       {'color': '#ab77ff', 'position': 0.4},
       {'color': '#ff0040', 'position': 1}
    ]
    ```
    """
    # select value
    for i in range(len(gradient_map) - 1):
        if value <= gradient_map[i + 1]["position"]:
            color1 = gradient_map[i]["color"]
            color2 = gradient_map[i + 1]["color"]
            position1 = gradient_map[i]["position"]
            position2 = gradient_map[i + 1]["position"]
            break

    # interpolated color based on the value
    ratio = (value - position1) / (position2 - position1)
    r = int(color1[0] + (color2[0] - color1[0]) * ratio)
    g = int(color1[1] + (color2[1] - color1[1]) * ratio)
    b = int(color1[2] + (color2[2] - color1[2]) * ratio)

    return r, g, b


def check_if_present(dock_device: RazerDevice, mouse_device: RazerDevice) -> bool:
    """## Returns ``True`` if both the mouse and the dock are present"""
    # * DPI values that are (0, 0) imply that the dongle is connected but the mouse is not.
    if not (bool(mouse_device) and bool(dock_device)) or getattr(mouse_device, "dpi", None) == (0, 0):
        if dock_device:
            if dock_device.fx.effect != "breathSingle":
                dock_device.fx.breath_single(50, 70, 255)
        print(f"(One or more) required devices is not present:\n({mouse_device=}, {dock_device=})")
        return False
    return True


if __name__ == "__main__":
    for _init_try in range(FULL_RETRIES_AMOUT):
        dock_device = None
        mouse_device = None

        devman = openrazer.client.DeviceManager()
        devman.sync_effects = False
        devices: list[RazerDevice] = devman.devices

        print(devices)

        for razer_device in devices:
            if "Dock" in razer_device.name:
                dock_device = razer_device
            if any(_mouse_name in razer_device.name for _mouse_name in CHROMA_DOCK_MICE_NAMES):
                mouse_device = razer_device
            if not razer_device.has('battery'):
                continue

            print(f"{razer_device.name} is at {razer_device.battery_level}% battery.")

        _present = check_if_present(dock_device, mouse_device)
        if not _present:
            # ! I added the this logic because the mouse is not always connected upon OS startup
            for _ in range(QUICK_RETRIES_AMOUT):
                time.sleep(5)
                _present = check_if_present(dock_device, mouse_device)
                if _present:
                    break
            else:
                if _init_try == FULL_RETRIES_AMOUT - 1:
                    print(f"Startup failed, retried {QUICK_RETRIES_AMOUT * FULL_RETRIES_AMOUT} times")
                    if dock_device:
                        dock_device.fx.static(139, 0, 181)
                    exit(1)
                else:
                    print(f"{QUICK_RETRIES_AMOUT} tries failed...")
                    continue

        dock_device.brightness = 100.0

        while _present:
            if not check_if_present(dock_device, mouse_device):
                print("Retrying...")
                time.sleep(5)
                continue

            _gradient_color = pick_gradient_color(mouse_device.battery_level, GRADIENT_RGB)
            print(f"Dock: {_gradient_color}; Mouse: {mouse_device.battery_level}%")
            dock_device.fx.static(*_gradient_color)
            time.sleep(POLL_DELAY_CONST)



# SPDX-License-Identifier: GPL-2.0-or-later

"""
Mousemat class
"""
from openrazer_daemon.hardware.device_base import RazerDeviceBrightnessSuspend as __RazerDeviceBrightnessSuspend


class RazerFirefly(__RazerDeviceBrightnessSuspend):
    """
    Class for the Razer Firefly (2013)
    """
    USB_VID = 0x1532
    USB_PID = 0x0C00
    HAS_MATRIX = True
    MATRIX_DIMS = [1, 15]
    METHODS = ['get_device_type_mousemat', 'set_wave_effect', 'set_static_effect', 'set_spectrum_effect',
               'set_reactive_effect', 'set_none_effect', 'set_breath_random_effect', 'set_breath_single_effect', 'set_breath_dual_effect',
               'set_custom_effect', 'set_key_row', 'trigger_reactive_effect']

    DEVICE_IMAGE = "https://assets.razerzone.com/eeimages/support/products/594/594_firefly_500x500.png"


class RazerFireflyV2(__RazerDeviceBrightnessSuspend):
    """
    Class for the Razer Firefly V2
    """
    USB_VID = 0x1532
    USB_PID = 0x0C04
    HAS_MATRIX = True
    MATRIX_DIMS = [1, 19]
    METHODS = ['get_device_type_mousemat', 'set_wave_effect', 'set_static_effect', 'set_spectrum_effect',
               'set_none_effect', 'set_breath_random_effect', 'set_breath_single_effect', 'set_breath_dual_effect',
               'set_custom_effect', 'set_key_row', 'set_reactive_effect', 'trigger_reactive_effect']

    DEVICE_IMAGE = "https://assets.razerzone.com/eeimages/support/products/1594/1594_firefly_v2.png"


class RazerFireflyV2Pro(__RazerDeviceBrightnessSuspend):
    """
    Class for the Razer Firefly V2 Pro
    """
    USB_VID = 0x1532
    USB_PID = 0x0C08
    HAS_MATRIX = True
    MATRIX_DIMS = [1, 17]
    METHODS = ['get_device_type_mousemat', 'set_wave_effect', 'set_static_effect', 'set_spectrum_effect',
               'set_none_effect', 'set_breath_random_effect', 'set_breath_single_effect', 'set_breath_dual_effect',
               'set_custom_effect', 'set_key_row', 'set_reactive_effect', 'trigger_reactive_effect']

    DEVICE_IMAGE = "https://dl.razerzone.com/src2/14141/14141-1-en-v1.png"


class RazerFireflyHyperflux(__RazerDeviceBrightnessSuspend):
    """
    Class for the Razer Firefly Hyperflux (2018)
    """
    USB_VID = 0x1532
    USB_PID = 0x0068
    HAS_MATRIX = True
    MATRIX_DIMS = [1, 17]
    METHODS = ['get_device_type_mousemat', 'set_static_effect', 'set_spectrum_effect', 'set_key_row', 'set_custom_effect',
               'set_none_effect', 'set_breath_random_effect', 'set_breath_single_effect', 'set_breath_dual_effect',
               'set_reactive_effect', 'trigger_reactive_effect']

    DEVICE_IMAGE = "https://assets.razerzone.com/eeimages/support/products/594/594_firefly_500x500.png"


class RazerGoliathus(__RazerDeviceBrightnessSuspend):
    """
    Class for the Razer Goliathus (2018)
    """
    USB_VID = 0x1532
    USB_PID = 0x0C01
    HAS_MATRIX = True
    MATRIX_DIMS = [1, 1]
    METHODS = ['get_device_type_mousemat', 'set_static_effect', 'set_spectrum_effect',
               'set_none_effect', 'set_breath_random_effect', 'set_breath_single_effect', 'set_breath_dual_effect',
               'set_custom_effect', 'set_key_row', 'set_reactive_effect', 'trigger_reactive_effect']

    DEVICE_IMAGE = "https://assets.razerzone.com/eeimages/support/products/1394/1394_goliathus_chroma.png"


class RazerGoliathusExtended(__RazerDeviceBrightnessSuspend):
    """
    Class for the Razer Goliathus Extended (2018)
    """
    USB_VID = 0x1532
    USB_PID = 0x0C02
    HAS_MATRIX = True
    MATRIX_DIMS = [1, 1]
    METHODS = ['get_device_type_mousemat', 'set_static_effect', 'set_spectrum_effect',
               'set_none_effect', 'set_breath_random_effect', 'set_breath_single_effect', 'set_breath_dual_effect',
               'set_custom_effect', 'set_key_row', 'set_reactive_effect', 'trigger_reactive_effect']

    DEVICE_IMAGE = "https://assets.razerzone.com/eeimages/support/products/1395/1395_goliathusextended.png"


class RazerGoliathusChroma3XL(__RazerDeviceBrightnessSuspend):
    """
    Class for the Razer Goliathus Chroma 3XL
    """
    USB_VID = 0x1532
    USB_PID = 0x0C06
    HAS_MATRIX = True
    MATRIX_DIMS = [1, 1]
    METHODS = ['get_device_type_mousemat', 'set_static_effect', 'set_spectrum_effect',
               'set_none_effect', 'set_breath_random_effect', 'set_breath_single_effect', 'set_breath_dual_effect',
               'set_custom_effect', 'set_key_row', 'set_reactive_effect', 'trigger_reactive_effect']

    DEVICE_IMAGE = "https://assets.razerzone.com/eeimages/support/products/1395/1395_goliathusextended.png"


class RazerStriderChroma(__RazerDeviceBrightnessSuspend):
    """
    Class for the Razer Strider Chroma
    """
    USB_VID = 0x1532
    USB_PID = 0x0C05
    HAS_MATRIX = True
    MATRIX_DIMS = [1, 19]
    METHODS = ['get_device_type_mousemat', 'set_wave_effect', 'set_static_effect', 'set_spectrum_effect',
               'set_none_effect', 'set_breath_random_effect', 'set_breath_single_effect', 'set_breath_dual_effect',
               'set_custom_effect', 'set_key_row', 'set_reactive_effect', 'trigger_reactive_effect']

    DEVICE_IMAGE = "https://assets2.razerzone.com/images/pnx.assets/73158705dfbee8e688bab3c7ae99c6d3/razer-strider-chroma-extended-size-1920x700.webp"



import sys
import argparse

import openrazer.client

choices = ('get', 'on', 'off')
parser = argparse.ArgumentParser()
parser.add_argument('action', choices=choices)
args = parser.parse_args()

device_manager = openrazer.client.DeviceManager()
keyboard = None

for device in device_manager.devices:
    if device.type == 'keyboard':
        keyboard = device
        break
else:
    print("Could not find suitable keyboard", file=sys.stderr)
    sys.exit(1)


# Check keyboard has game_mode
if not keyboard.has('game_mode_led'):
    print("Keyboard doesn't have game mode", file=sys.stderr)
    sys.exit(1)


if args.action == 'get':
    if keyboard.game_mode_led:
        print("Game mode LED: On")
    else:
        print("Game mode LED: Off")
elif args.action == 'on':
    keyboard.game_mode_led = True
    print("Turned game mode LED on")
else:
    keyboard.game_mode_led = False
    print("Turned game mode LED off")



# SPDX-License-Identifier: GPL-2.0-or-later

import dbus
from openrazer.client.devices import RazerDevice, BaseDeviceFactory
from openrazer.client.devices.mousemat import RazerMousemat
from openrazer.client.devices.keyboard import RazerKeyboard
from openrazer.client.devices.mice import RazerMouse


DEVICE_MAP = {
    'mousemat': RazerMousemat,
    'keyboard': RazerKeyboard,
    'mouse': RazerMouse,
    'keypad': RazerKeyboard,
}


class RazerDeviceFactory(BaseDeviceFactory):
    """
    Simple factory to return an object for a given device

    """
    @staticmethod
    def get_device(serial: str) -> RazerDevice:
        """
        Factory for turning a serial into a class

        Device factory, will return a class fit for the device in question. The DEVICE_MAP mapping above
        can contain a device_type => DeviceClass or DeviceFactory, this allows us to specify raw device classes
        if there is only one model (like Firefly) or a factory for the keyboards (so we can differentiate between
        old blackwidows and chromas). If the device is not in the device mapping then the factory will default
        to a raw RazerDevice.

        :param serial: Device serial
        :type serial: str

        :return: RazerDevice object (or subclass)
        :rtype: RazerDevice
        """
        session_bus = dbus.SessionBus()
        daemon_dbus = session_bus.get_object("org.razer", "/org/razer/device/{0}".format(serial))
        device_dbus = dbus.Interface(daemon_dbus, "razer.device.misc")

        device_type = device_dbus.getDeviceType()

        device_class = DEVICE_MAP.get(device_type, RazerDevice)

        return device_class(serial)



#!/usr/bin/python3
import openrazer.client
import glob

daemon_test_dir = "/tmp/daemon_test"
devmgr = openrazer.client.DeviceManager()
passed = True


def _test_failed(name, msg):
    global passed
    passed = False
    print(f"\nFAILED: {name}\n       ", msg)


def test_sanity_check_matrix_capabilities(d):
    if d.has("lighting_led_matrix"):
        d.fx.advanced.matrix[0, 0] = [0, 255, 0]
        try:
            d.fx.advanced.draw()
        except Exception as e:
            _test_failed(d.name, e)


def test_wired_wireless_naming():
    # Enforce the (Wired) or (Wireless) suffix naming convention when devices have multiple PIDs
    device_names = list(map(lambda device: device.name, devmgr.devices))

    for name in device_names:
        if name.find("(Wired)") != -1:
            striped_name = name.replace("(Wired)", "").strip()
            if striped_name in device_names:
                _test_failed(striped_name, "Naming pattern is inconsistent. Append \"(Wireless)\" for the other USB PID.")

        elif name.find("(Wireless)") != -1:
            striped_name = name.replace("(Wireless)", "").strip()
            if striped_name in device_names:
                _test_failed(striped_name, "Naming pattern is inconsistent. Append \"(Wired)\" for the other USB PID.")


def test_sysfs_consistency(d):
    # Check sysfs files exist for capabilities
    vid = str(hex(d._vid))[2:].upper().rjust(4, '0')
    pid = str(hex(d._pid))[2:].upper().rjust(4, '0')

    def check_sysfs(capability: str, sysfs_name: str):
        """
        Check the device has either the given pylib capability for the
        given sysfs name, and vice versa.
        """
        try:
            expected_path = glob.glob(f"{daemon_test_dir}/*:{vid}:{pid}*/{sysfs_name}", recursive=True)[0]
        except IndexError:
            expected_path = ""

        if d.has(capability) and not glob.glob(expected_path, recursive=True):
            _test_failed(d.name, str(hex(d._pid)) + " Has capability '{}' but no sysfs file: '{}'".format(capability, sysfs_name))

        if not d.has(capability) and glob.glob(expected_path, recursive=True):
            _test_failed(d.name, str(hex(d._pid)) + " Has sysfs '{}' but no capability set: '{}'".format(sysfs_name, capability))

    def check_any_sysfs(capabilities: list[str], sysfs_names: list[str]):
        """
        Check the device has one of these sysfs names for one of these pylib capabilities.
        """
        found_sysfs = []
        found_capability = []

        for sysfs_name in sysfs_names:
            if glob.glob(f"{daemon_test_dir}/*:{vid}:{pid}*/{sysfs_name}", recursive=True):
                found_sysfs.append(sysfs_name)

        for capability in capabilities:
            if d.has(capability):
                found_capability.append(capability)

        # Skip Razer devices that have non-RGB effects
        if not found_capability and ("logo_led_effect" in sysfs_names or "scroll_led_effect" in sysfs_names):
            return

        if not found_capability and found_sysfs:
            _test_failed(d.name, str(hex(d._pid)) + " Has one of these sysfs {} but none of these capabilities: {}".format(sysfs_names, capabilities))

        if found_capability and not found_sysfs:
            _test_failed(d.name, str(hex(d._pid)) + " Has one of these capabilities {} but none of these sysfs files: {}".format(capabilities, sysfs_names))

    # check_sysfs("battery", "charge_effect") # FIXME this is not correct, as per PR comments
    check_sysfs("set_low_battery_threshold", "charge_low_threshold")  # deprecated
    check_sysfs("low_battery_threshold", "charge_low_threshold")
    check_sysfs("set_idle_time", "device_idle_time")  # deprecated
    check_sysfs("idle_time", "device_idle_time")

    check_sysfs("battery", "charge_level")
    check_sysfs("battery", "charge_status")
    check_sysfs("brightness", "matrix_brightness")
    check_sysfs("dpi", "dpi")
    check_sysfs("dpi_stages", "dpi_stages")
    check_sysfs("firmware_version", "firmware_version")
    check_sysfs("game_mode_led", "game_led_state")
    check_sysfs("macro_mode_led", "macro_led_state")
    check_sysfs("macro_mode_led_effect", "macro_led_effect")
    check_sysfs("poll_rate", "poll_rate")
    check_sysfs("serial", "device_serial")
    check_sysfs("type", "device_type")

    if d.type == "keyboard":
        check_sysfs("keyboard_layout", "kbd_layout")

    # Devices with a matrix defined should have a suitable "custom" sysfs. Krakens are a special case.
    if d._matrix_dimensions != (-1, -1) and d.name.find("Kraken") == -1:
        check_sysfs("lighting_led_matrix", "matrix_custom_frame")
        check_sysfs("lighting_led_matrix", "matrix_effect_custom")

    check_sysfs("lighting_breath_single", "matrix_effect_breath")
    check_sysfs("lighting_none", "matrix_effect_none")
    check_sysfs("lighting_reactive", "matrix_effect_reactive")
    check_sysfs("reactive_trigger", "matrix_reactive_trigger")
    check_sysfs("lighting_spectrum", "matrix_effect_spectrum")
    check_any_sysfs(["lighting_starlight_random", "lighting_starlight_single", "lighting_starlight_dual"], ["matrix_effect_starlight"])
    # Ignore devices that have mono-color razer.device.lighting.bw2013.setStatic()
    if d._pid not in [0x010d, 0x010e, 0x0113, 0x0118, 0x011a, 0x011b, 0x011c, 0x0202]:
        check_sysfs("lighting_static", "matrix_effect_static")
    check_sysfs("lighting_wave", "matrix_effect_wave")
    check_sysfs("lighting_wheel", "matrix_effect_wheel")
    check_sysfs("lighting_pulsate", "matrix_effect_pulsate")
    check_sysfs("lighting_blinking", "matrix_effect_blinking")

    check_sysfs("lighting_backlight_active", "backlight_led_state")

    for prefix in ["logo", "scroll", "left", "right", "charging", "fast_charging", "fully_charged"]:
        check_sysfs(f"lighting_{prefix}_brightness", f"{prefix}_led_brightness")
        check_sysfs(f"lighting_{prefix}_active", f"{prefix}_led_state")

        for effect in ["none", "reactive", "spectrum", "static", "wave"]:
            # There are two implementations with different sysfs files
            check_any_sysfs([f"lighting_{prefix}_{effect}"], [f"{prefix}_matrix_effect_{effect}", f"{prefix}_led_effect"])

        check_any_sysfs([f"lighting_{prefix}_breath_single", f"lighting_{prefix}_breath_mono"], [f"{prefix}_matrix_effect_breath"])


def test_ripple_capable(d):
    # Check that the device has a matrix for a software ripple effect
    if d.has("lighting_ripple") or d.has("lighting_ripple_random"):
        if d._matrix_dimensions == (-1, -1):
            _test_failed(d.name, "Cannot have a ripple capability without a matrix")


for d in devmgr.devices:
    test_sanity_check_matrix_capabilities(d)
    test_sysfs_consistency(d)
    test_ripple_capable(d)
test_wired_wireless_naming()

if not passed:
    exit(1)



# SPDX-License-Identifier: GPL-2.0-or-later

import unittest
import unittest.mock

import openrazer_daemon.misc.effect_sync

# msg type = effect, arg = orig_device, arg = effect_name, arg.. = arg..
MSG1 = ('effect', None, 'setBrightness', 255)

# Static effect message from blackwidow chroma
MSG2 = ('effect', None, 'setStatic', 255, 255, 0)
# Static effect message from blackwidow standard
MSG3 = ('effect', None, 'setStatic')
# Breathing effect message from blackwidow chroma
MSG4 = ('effect', None, 'setBreathSingle', 255, 255, 0)
# Pulsate effect message from blackwidow standard
MSG5 = ('effect', None, 'setPulsate')


def logger_mock(*args):
    return unittest.mock.MagicMock()


class DummyHardwareDevice(object):
    def __init__(self):
        self.observer_list = []

        self.disable_notify = None

        self.effect_call = None

    def register_observer(self, obs):
        if obs not in self.observer_list:
            self.observer_list.append(obs)

    def remove_observer(self, obs):
        if obs in self.observer_list:
            self.observer_list.remove(obs)

    # Effect functions
    def setBrightness(self, brightness):
        self.effect_call = ('setBrightness', brightness)

    def setStatic(self):
        raise Exception("test")


class DummyHardwareBlackWidowStandard(DummyHardwareDevice):
    def setStatic(self):
        self.effect_call = ('setStatic',)

    def setPulsate(self):
        self.effect_call = ('setPulsate',)


class DummyHardwareBlackWidowChroma(DummyHardwareDevice):
    def setStatic(self, red, green, blue):
        self.effect_call = ('setStatic', red, green, blue)

    def setBreathSingle(self, red, green, blue):
        self.effect_call = ('setBreathSingle', red, green, blue)


class EffectSyncTest(unittest.TestCase):
    @unittest.mock.patch('openrazer_daemon.misc.effect_sync.logging.getLogger', logger_mock)
    def setUp(self):
        self.hardware_device = DummyHardwareDevice()
        self.effect_sync = openrazer_daemon.misc.effect_sync.EffectSync(self.hardware_device, 1)

    def test_observers(self):

        self.assertIn(self.effect_sync, self.hardware_device.observer_list)

        # Remove observers
        self.effect_sync.close()

        self.assertEqual(len(self.hardware_device.observer_list), 0)

    def test_get_num_arguments(self):

        def func_2_args(x, y): return x + y

        num_args = self.effect_sync.get_num_arguments(func_2_args)

        self.assertEqual(num_args, 2)

    def test_notify_invalid_message(self):
        self.effect_sync.notify("test")

        self.assertTrue(self.effect_sync._logger.warning.called)

    def test_notify_message_from_other(self):
        # Patch run_effect
        self.effect_sync.run_effect = unittest.mock.MagicMock()

        self.effect_sync.notify(MSG1)

        self.assertTrue(self.effect_sync.run_effect.called)

        new_msg = self.effect_sync.run_effect.call_args_list[0][0]
        # Check function is called run_effect('setBrightness', 255)
        self.assertEqual(new_msg[0], MSG1[2])
        self.assertEqual(new_msg[1], MSG1[3])

    def test_notify_run_effect(self):
        self.effect_sync.notify(MSG1)

        self.assertEqual(self.hardware_device.effect_call[0], MSG1[2])
        self.assertEqual(self.hardware_device.effect_call[1], MSG1[3])

        self.assertIsNotNone(self.hardware_device.disable_notify)
        self.assertFalse(self.hardware_device.disable_notify)

    def test_notify_run_effect_edge_case_1(self):
        # Set the parent to a blackwidow (non chroma)
        self.hardware_device = DummyHardwareBlackWidowStandard()
        self.effect_sync._parent = self.hardware_device
        self.hardware_device.register_observer(self.effect_sync)

        self.effect_sync.notify(MSG2)

        self.assertEqual(self.hardware_device.effect_call[0], MSG2[2])

    def test_notify_run_effect_edge_case_2(self):
        # Set the parent to a blackwidow chroma
        self.hardware_device = DummyHardwareBlackWidowChroma()
        self.effect_sync._parent = self.hardware_device
        self.hardware_device.register_observer(self.effect_sync)

        self.effect_sync.notify(MSG3)

        self.assertTupleEqual(self.hardware_device.effect_call, ('setStatic', 0, 255, 0))

    def test_notify_run_effect_edge_case_3(self):
        # Set the parent to a blackwidow (non chroma)
        self.hardware_device = DummyHardwareBlackWidowStandard()
        self.effect_sync._parent = self.hardware_device
        self.hardware_device.register_observer(self.effect_sync)

        self.effect_sync.notify(MSG4)
        # Send setBreath but converted it to setPulsate
        self.assertEqual(self.hardware_device.effect_call[0], 'setPulsate')

    def test_notify_run_effect_edge_case_4(self):
        # Set the parent to a blackwidow chroma
        self.hardware_device = DummyHardwareBlackWidowChroma()
        self.effect_sync._parent = self.hardware_device
        self.hardware_device.register_observer(self.effect_sync)

        self.effect_sync.notify(MSG5)

        self.assertTupleEqual(self.hardware_device.effect_call, ('setBreathSingle', 0, 255, 0))

    def test_notify_run_effect_edge_case_5(self):
        # Exception of standard setStatic() should be caught
        self.effect_sync.notify(MSG3)

        # Logger should have called .exception
        self.assertTrue(self.effect_sync._logger.exception.called)



#!/usr/bin/python3
import argparse
import glob
import os
import struct
import sys


def clamp_to_min_max(value, min, max):
    if value > max:
        value = max
    elif value < min:
        value = min
    return value


def clamp_to_u8(value):
    return clamp_to_min_max(value, 0, 255)


def parse_args():
    parser = argparse.ArgumentParser(description="Set the mouse DPI")
    parser.add_argument('-d', '--device', type=str, help="Device string like \"0003:1532:0045.000C\"")

    parser.add_argument('--dpi_x', required=True, type=int, help="DPI on the X axis (100, 30000)")
    parser.add_argument('--dpi_y', type=int, help="DPI on the Y axis (if omitted, dpi_x is used), (100, 30000)")

    args = parser.parse_args()
    return args


def run():
    args = parse_args()

    if args.device is None:
        mouse_dirs = glob.glob(os.path.join('/sys/bus/hid/drivers/razermouse/', "*:*:*.*"))

        if len(mouse_dirs) > 1:
            print("Multiple mouse directories found. Rerun with -d", file=sys.stderr)
            sys.exit(1)
        if len(mouse_dirs) < 1:
            print("No mouse directories found. Make sure the driver is binded", file=sys.stderr)
            sys.exit(1)

        mouse_dir = mouse_dirs[0]
    else:
        mouse_dir = os.path.join('/sys/bus/hid/drivers/razermouse/', args.device)

    if not os.path.isdir(mouse_dir):
        print("Multiple mouse directories found. Rerun with -d", file=sys.stderr)
        sys.exit(1)

    dpi_x = clamp_to_min_max(args.dpi_x, 100, 30000)
    if args.dpi_y is None:
        byte_string = struct.pack(">H", dpi_x)
    else:
        dpi_y = clamp_to_min_max(args.dpi_y, 100, 30000)
        byte_string = struct.pack(">HH", dpi_x, dpi_y)

    set_mouse_dpi_filepath = os.path.join(mouse_dir, "dpi")
    with open(set_mouse_dpi_filepath, 'wb') as set_mouse_dpi_file:
        set_mouse_dpi_file.write(byte_string)
    print("Done")


if __name__ == '__main__':
    run()



# SPDX-License-Identifier: GPL-2.0-or-later

"""
Screensaver class which watches dbus signals to see if screensaver is active
"""
import logging
import dbus
import dbus.exceptions

DBUS_SCREENSAVER_INTERFACES = (
    'org.cinnamon.ScreenSaver',
    'org.freedesktop.ScreenSaver',
    'org.gnome.ScreenSaver',
    'org.mate.ScreenSaver',
    'org.xfce.ScreenSaver',
)


class ScreensaverMonitor(object):
    """
    Simple class for monitoring signals on the Session Bus
    """

    def __init__(self, parent):
        self._logger = logging.getLogger('razer.screensaver')
        self._logger.info("Initialising DBus Screensaver Monitor")

        self._parent = parent
        self._monitoring = True
        self._active = None

        self._dbus_instances = []
        # Get session bus
        bus = dbus.SessionBus()
        # Loop through and monitor the signals
        for screensaver_interface in DBUS_SCREENSAVER_INTERFACES:
            bus.add_signal_receiver(self.signal_callback, dbus_interface=screensaver_interface, signal_name='ActiveChanged')

    @property
    def monitoring(self):
        """
        Monitoring property, if true then suspend/resume will be actioned.

        :return: If monitoring
        :rtype: bool
        """
        return self._monitoring

    @monitoring.setter
    def monitoring(self, value):
        """
        Monitoring property setter, if true then suspend/resume will be actioned.

        :param value: If monitoring
        :type: bool
        """
        self._monitoring = bool(value)

    def suspend(self):
        """
        Suspend the device
        """
        self._logger.debug("Received screensaver active signal")
        self._parent.suspend_devices()

    def resume(self):
        """
        Resume the device
        """
        self._logger.debug("Received screensaver inactive signal")
        self._parent.resume_devices()

    def signal_callback(self, active):
        """
        Called by DBus when a signal is found

        :param active: If the screensaver is active
        :type active: dbus.Boolean
        """
        active = bool(active)

        if self.monitoring:
            if active:
                # Only trigger once per state change
                if self._active is None or not self._active:
                    self._active = active
                    self.suspend()
            else:
                if self._active is None or self._active:
                    self._active = active
                    self.resume()



"""
Example of changing the colour of all Razer devices based on the CPU
temperature
"""
import colorsys
import time

import openrazer.client

# These settings are nice if your CPU temperature is a bit less than 30°C
# at idle and a bit more than 60°C at full load, tune them for your needs
tGreen = 30
tRed = 60

tRange = 3 * (tRed - tGreen)
colour0 = tRed / tRange
tRange = 1000 * tRange

devman = openrazer.client.DeviceManager()
devman.sync_effects = False
devices = devman.devices

if not devices:
    print("No devices connected")
    exit(1)

lastTemp = 0

while True:
    # For my computer it's hwmon2/temp1_input but it may be different for you
    # cat /sys/class/hwmon/hwmonX/tempY_input can help you (with X between 0
    # and 2 and Y between 0 and 12, the result is in millidegrees).
    # you will need lm-sensors for this to work.
    with open("/sys/class/hwmon/hwmon2/temp1_input", "r") as f:
        temp = f.readline()

    shortTemp = temp[:2]

    # Now we can change the colour each time temperature changes
    if shortTemp != lastTemp:
        # We convert the temperature (millidegrees) to the hue
        # (HSV, hue between 0 and 1)
        hue = (colour0 - float(temp) / tRange) % 1

        # The colours list RGB with R,G,B between 0 and 1
        colour = colorsys.hsv_to_rgb(hue, 1, 1)

        # Needs to be an integer (0-255, e.g. 1 == 255)
        colour = list(map(lambda x: (int(255 * x)), colour))

        # Set effect on device
        for device in devices:
            # A device may have multiple zones
            if device.has("lighting_static"):
                device.fx.static(colour[0], colour[1], colour[2])
            if device.has("lighting_backlight"):
                device.fx.misc.backlight.static(colour[0], colour[1], colour[2])
            if device.has("lighting_logo_static"):
                device.fx.misc.logo.static(colour[0], colour[1], colour[2])
            if device.has("lighting_scroll_static"):
                device.fx.misc.scroll_wheel.static(colour[0], colour[1], colour[2])
            if device.has("lighting_left_static"):
                device.fx.misc.left.static(colour[0], colour[1], colour[2])
            if device.has("lighting_right_static"):
                device.fx.misc.right.static(colour[0], colour[1], colour[2])

        lastTemp = shortTemp

    time.sleep(1)



# SPDX-License-Identifier: GPL-2.0-or-later

"""
Hardware collection
"""
import os
from openrazer_daemon.hardware.device_base import RazerDevice

# Hack to get a list of hardware modules to import
HARDWARE_MODULES = ['openrazer_daemon.hardware.' + os.path.splitext(hw_file)[0] for hw_file in os.listdir(os.path.dirname(__file__)) if hw_file not in ('device_base.py', '__init__.py') and hw_file.endswith('.py')]

# List of classes to exclude from the class finding
EXCLUDED_CLASSES = ('RazerDevice', 'RazerDeviceBrightnessSuspend')


def get_device_classes():
    """
    Get a list of hardware classes

    :return: List of RazerDevice subclasses
    :rtype: list of callable
    """
    classes = []

    for hw_module in HARDWARE_MODULES:
        imported_module = __import__(hw_module, globals=globals(), locals=locals(), fromlist=['*'], level=0)

        for class_name, class_instance in imported_module.__dict__.items():
            if class_name in EXCLUDED_CLASSES or class_name.startswith('_') or not isinstance(class_instance, type):
                continue

            classes.append(class_instance)

    return sorted(classes, key=lambda cls: cls.__name__)
